package authorizationheader

import (
	"crypto/x509"
	"log"
	"net/http"
	"sync"
	"time"

	"github.com/craigfurman/herottp"
)

type HTTPClient interface {
	Do(req *http.Request) (*http.Response, error)
}

type ClientTokenAuthHeaderBuilder struct {
	uaaURL            string
	clientID          string
	clientSecret      string
	httpClient        HTTPClient
	cachedToken       string
	cachedTokenExpiry time.Time
	tokenLock         *sync.Mutex
}

type ObtainTokenResponse struct {
	Token            string `json:"access_token"`
	ExpiresInSeconds int    `json:"expires_in"`
}

var MinimumRemainingValidity time.Duration = time.Second * 10

func NewClientTokenAuthHeaderBuilder(uaaURL, clientID, clientSecret string, disableSSLCertVerification bool, trustedCertPEM []byte) (*ClientTokenAuthHeaderBuilder, error) {
	rootCAs, err := x509.SystemCertPool()
	if err != nil {
		return nil, err
	}
	rootCAs.AppendCertsFromPEM(trustedCertPEM)

	return &ClientTokenAuthHeaderBuilder{
		uaaURL:       uaaURL,
		clientID:     clientID,
		clientSecret: clientSecret,
		httpClient: herottp.New(herottp.Config{
			DisableTLSCertificateVerification: disableSSLCertVerification,
			RootCAs: rootCAs,
			Timeout: 30 * time.Second,
		}),
		tokenLock: new(sync.Mutex),
	}, nil
}

func (hb *ClientTokenAuthHeaderBuilder) Build(logger *log.Logger) (string, error) {
	hb.tokenLock.Lock()
	defer hb.tokenLock.Unlock()

	var err error
	hb.cachedToken, hb.cachedTokenExpiry, err = getValidToken(hb.cachedToken, hb.cachedTokenExpiry, logger, hb.obtainToken)
	if err != nil {
		return "", err
	}

	return bearerToken(hb.cachedToken), nil
}

func (hb *ClientTokenAuthHeaderBuilder) obtainToken(logger *log.Logger) (string, int, error) {
	requestBody := "grant_type=client_credentials"

	request, err := buildTokenRequest(requestBody, hb.uaaURL, hb.clientID, hb.clientSecret)
	if err != nil {
		return "", 0, err
	}

	return doObtainTokenRequest(hb.httpClient, logger, request)
}
