package authorizationheader

import (
	"crypto/x509"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/craigfurman/herottp"
)

type UserTokenAuthHeaderBuilder struct {
	uaaURL            string
	clientID          string
	clientSecret      string
	username          string
	password          string
	httpClient        HTTPClient
	cachedToken       string
	cachedTokenExpiry time.Time
	tokenLock         *sync.Mutex
}

func NewUserTokenAuthHeaderBuilder(uaaURL, clientID, clientSecret, username, password string, disableSSLCertVerification bool, trustedCertPEM []byte) (*UserTokenAuthHeaderBuilder, error) {
	rootCAs, err := x509.SystemCertPool()
	if err != nil {
		return nil, err
	}
	rootCAs.AppendCertsFromPEM(trustedCertPEM)
	return &UserTokenAuthHeaderBuilder{
		uaaURL:       uaaURL,
		clientID:     clientID,
		clientSecret: clientSecret,
		username:     username,
		password:     password,
		httpClient: herottp.New(herottp.Config{
			DisableTLSCertificateVerification: disableSSLCertVerification,
			RootCAs: rootCAs,
			Timeout: 30 * time.Second,
		}),
		tokenLock: new(sync.Mutex),
	}, nil
}

func (hb *UserTokenAuthHeaderBuilder) Build(logger *log.Logger) (string, error) {
	hb.tokenLock.Lock()
	defer hb.tokenLock.Unlock()

	var err error
	hb.cachedToken, hb.cachedTokenExpiry, err = getValidToken(hb.cachedToken, hb.cachedTokenExpiry, logger, hb.obtainToken)
	if err != nil {
		return "", err
	}

	return bearerToken(hb.cachedToken), nil
}

func (hb *UserTokenAuthHeaderBuilder) obtainToken(logger *log.Logger) (string, int, error) {
	requestBody := fmt.Sprintf("grant_type=password&username=%s&password=%s", hb.username, hb.password)

	request, err := buildTokenRequest(requestBody, hb.uaaURL, hb.clientID, hb.clientSecret)
	if err != nil {
		return "", 0, err
	}

	return doObtainTokenRequest(hb.httpClient, logger, request)
}
