package boshclient

import (
	"bytes"
	"crypto/x509"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net/http"
	"strconv"
	"strings"
	"time"

	"github.com/craigfurman/herottp"
	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
)

type Client struct {
	boshURL string

	BoshPollingInterval time.Duration

	authHeaderBuilder authorizationheader.AuthHeaderBuilder
	httpClient        HTTPClient
}

type HTTPClient interface {
	Do(req *http.Request) (*http.Response, error)
}

func New(boshURL string, authHeaderBuilder authorizationheader.AuthHeaderBuilder, disableSSLCertVerification bool, trustedCertPEM []byte) (*Client, error) {
	rootCAs, err := x509.SystemCertPool()
	if err != nil {
		return nil, err
	}
	rootCAs.AppendCertsFromPEM(trustedCertPEM)
	return &Client{
		boshURL:           boshURL,
		authHeaderBuilder: authHeaderBuilder,
		httpClient: herottp.New(herottp.Config{
			NoFollowRedirect:                  true,
			DisableTLSCertificateVerification: disableSSLCertVerification,
			RootCAs: rootCAs,
			Timeout: 30 * time.Second,
		}),
		BoshPollingInterval: 5,
	}, nil
}

type BoshInfo struct {
	Version string
}

type BoshTask struct {
	ID          int
	State       string
	Description string
	Result      string
}

type BoshDeployment struct {
	Name string
}

func (t BoshTask) ToLog() string {
	return fmt.Sprintf("{ID: %d, state: %q, description: %q, result: %q}", t.ID, t.State, t.Description, t.Result)
}

const (
	BoshTaskQueued     = "queued"
	BoshTaskProcessing = "processing"
	BoshTaskDone       = "done"
	BoshTaskError      = "error"
	BoshTaskCancelled  = "cancelled"
	BoshTaskCancelling = "cancelling"
	BoshTaskTimeout    = "timeout"
)

type DeploymentNotFoundError struct {
	error
}

type RequestError struct {
	error
}

func NewRequestError(e error) RequestError {
	return RequestError{e}
}

type unexpectedStatusError struct {
	expectedStatus int
	actualStatus   int
	responseBody   string
}

func (u unexpectedStatusError) Error() string {
	return fmt.Sprintf("expected status %d, was %d. Response Body: %s", u.expectedStatus, u.actualStatus, u.responseBody)
}

type resultExtractor func(*http.Response) error

func (c *Client) getDataFromBoshCheckingForErrors(url string, expectedStatus int, result interface{}, logger *log.Logger) error {
	request, err := prepareGet(url)
	if err != nil {
		return err
	}
	return c.getResultFromBoshCheckingForErrors(request, expectedStatus, decodeJson(result), logger)
}

func (c *Client) getTaskIdFromBoshCheckingForErrors(url string, expectedStatus int, logger *log.Logger) (int, error) {
	request, err := prepareGet(url)
	if err != nil {
		return 0, err
	}
	var taskId int
	err = c.getDeploymentResultFromBoshCheckingForErrors(request, expectedStatus, extractTaskId(&taskId), logger)
	return taskId, err
}

func (c *Client) getMultipleDataFromBoshCheckingForErrors(url string, expectedStatus int, result interface{}, resultReady func(), logger *log.Logger) error {
	request, err := prepareGet(url)
	if err != nil {
		return err
	}
	return c.getResultFromBoshCheckingForErrors(request, expectedStatus, decodeMultipleJson(result, resultReady), logger)
}

func (c *Client) postAndGetTaskIdFromBoshCheckingForErrors(url string, expectedStatus int, body []byte, contentType string, logger *log.Logger) (int, error) {
	request, err := preparePost(url, body, contentType)
	if err != nil {
		return 0, err
	}
	var taskId int
	err = c.getResultFromBoshCheckingForErrors(request, expectedStatus, extractTaskId(&taskId), logger)
	return taskId, err
}

func (c *Client) deleteAndGetTaskIdFromBoshCheckingForErrors(url string, expectedStatus int, logger *log.Logger) (int, error) {
	request, err := prepareDelete(url)
	if err != nil {
		return 0, err
	}
	var taskId int
	err = c.getDeploymentResultFromBoshCheckingForErrors(request, expectedStatus, extractTaskId(&taskId), logger)
	return taskId, err
}

func decodeJson(result interface{}) resultExtractor {
	return func(response *http.Response) error {
		return json.NewDecoder(response.Body).Decode(result)
	}
}

func decodeMultipleJson(result interface{}, resultReady func()) resultExtractor {
	return func(response *http.Response) error {
		decoder := json.NewDecoder(response.Body)
		for decoder.More() {
			if err := decoder.Decode(&result); err != nil {
				return err
			}
			resultReady()
		}
		return nil
	}
}

func extractTaskId(taskId *int) resultExtractor {
	return func(response *http.Response) error {
		var e error
		urlParts := strings.Split(response.Header.Get("location"), "/")
		*taskId, e = strconv.Atoi(urlParts[len(urlParts)-1])
		return e
	}
}

func taskIDFromRedirect(responseHeader http.Header) (int, error) {
	urlParts := strings.Split(responseHeader.Get("location"), "/")
	taskID, err := strconv.Atoi(urlParts[len(urlParts)-1])
	if err != nil {
		return 0, fmt.Errorf("error parsing task ID: %s", err)
	}

	return taskID, nil
}

func prepareGet(url string) (*http.Request, error) {
	return http.NewRequest("GET", url, nil)
}

func preparePost(url string, body []byte, contentType string) (*http.Request, error) {
	request, err := http.NewRequest("POST", url, bytes.NewReader(body))
	if err != nil {
		return nil, err
	}
	request.Header.Set("Content-Type", contentType)
	return request, nil
}

func prepareDelete(url string) (*http.Request, error) {
	return http.NewRequest("DELETE", url, nil)
}

func (c *Client) getDeploymentResultFromBoshCheckingForErrors(request *http.Request, expectedStatus int, handler resultExtractor, logger *log.Logger) error {
	err := c.getResultFromBoshCheckingForErrors(request, expectedStatus, handler, logger)
	switch err := err.(type) {
	case unexpectedStatusError:
		if err.actualStatus == http.StatusNotFound {
			return DeploymentNotFoundError{error: err}
		}
		return err
	default:
		return err
	}
}

func (c *Client) getResultFromBoshCheckingForErrors(request *http.Request, expectedStatus int, handler resultExtractor, logger *log.Logger) error {
	authHeader, err := c.authHeaderBuilder.Build(logger)
	if err != nil {
		return err
	}
	request.Header.Set("Authorization", authHeader)

	response, err := c.httpClient.Do(request)
	if err != nil {
		return NewRequestError(fmt.Errorf("error reaching bosh director: %s. Please make sure that properties.<broker-job>.bosh.url is correct and reachable.", err))
	}
	defer response.Body.Close()

	if response.StatusCode != expectedStatus {
		return unexpectedStatusErr(response, expectedStatus)
	}

	return handler(response)
}

func unexpectedStatusErr(response *http.Response, expectedStatus int) error {
	return unexpectedStatusError{
		expectedStatus: expectedStatus,
		actualStatus:   response.StatusCode,
		responseBody:   readBody(response.Body),
	}
}

func readBody(bodyReader io.ReadCloser) string {
	body, err := ioutil.ReadAll(bodyReader)
	if err == nil {
		return string(body)
	} else {
		return fmt.Sprintf("COULDN'T READ RESPONSE BODY: %s", err)
	}
}
