package boshclient_test

import (
	"errors"

	"gopkg.in/yaml.v2"

	authorizationheader "github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("deploying a bosh release", func() {
	const (
		taskID = 3
	)

	var (
		deploymentManifest = bosh.BoshManifest{
			Name:           "big deployment",
			Releases:       []bosh.Release{},
			Stemcells:      []bosh.Stemcell{},
			InstanceGroups: []bosh.InstanceGroup{},
		}
		deploymentManifestBytes []byte

		returnedTaskID int
		deployErr      error
	)

	BeforeEach(func() {
		var err error
		deploymentManifestBytes, err = yaml.Marshal(deploymentManifest)
		Expect(err).NotTo(HaveOccurred())
	})

	AfterEach(func() {
		director.Close()
	})

	Context("using a structured manifest", func() {
		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deploy().WithRawManifest(deploymentManifestBytes).RedirectsToTask(taskID),
				)
			})

			JustBeforeEach(func() {
				returnedTaskID, deployErr = c.Deploy(deploymentManifest, logger)
			})

			It("calls the authorization header builder", func() {
				Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
			})

			It("deploys release with specified name", func() {
				Expect(deployErr).NotTo(HaveOccurred())
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})
		})

		Context("when BOSH cannot deploy the release", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deploy().RedirectsToTask(taskID).Fails("because reasons"),
				)
			})

			JustBeforeEach(func() {
				returnedTaskID, deployErr = c.Deploy(deploymentManifest, logger)
			})

			It("returns error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("expected status 302, was 500.")))
			})
		})

		Context("when the BOSH director cannot be reached", func() {
			It("returns error", func() {
				c, err := boshclient.New("http://localhost", authorizationheader.NewBasicAuthHeaderBuilder("", ""), false, nil)
				Expect(err).NotTo(HaveOccurred())
				_, err = c.Deploy(deploymentManifest, logger)
				Expect(err).To(MatchError(ContainSubstring("error reaching bosh director")))
			})
		})

		Context("when the Authorization header cannot be generated", func() {
			BeforeEach(func() {
				authHeaderBuilder.BuildReturns("", errors.New("some-error"))
			})

			JustBeforeEach(func() {
				returnedTaskID, deployErr = c.Deploy(deploymentManifest, logger)
			})

			It("returns an error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("some-error")))
			})
		})
	})

	Context("using a raw manifest", func() {
		var deploymentManifestBytes []byte
		BeforeEach(func() {
			var err error
			deploymentManifestBytes, err = yaml.Marshal(deploymentManifest)
			Expect(err).NotTo(HaveOccurred())

			director.VerifyAndMock(
				mockbosh.Deploy().WithRawManifest(deploymentManifestBytes).RedirectsToTask(taskID),
			)
		})

		JustBeforeEach(func() {
			returnedTaskID, deployErr = c.DeployRaw(deploymentManifestBytes, logger)
		})

		Context("when bosh deploys the release successfully", func() {
			It("deploys release with specified name", func() {
				Expect(deployErr).NotTo(HaveOccurred())
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})
		})
	})

})
