package boshclient_test

import (
	"encoding/json"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"
)

var _ = Describe("deployments", func() {
	Context("gets deployments", func() {
		var (
			actualDeployments      []boshclient.BoshDeployment
			actualDeploymentsError error
		)

		JustBeforeEach(func() {
			actualDeployments, actualDeploymentsError = c.GetDeployments(logger)
		})

		Context("when bosh fetches the deployments successfully", func() {
			var expectedDeployments []boshclient.BoshDeployment

			BeforeEach(func() {
				expectedDeployments = []boshclient.BoshDeployment{
					{Name: "service-instance_monkeys"},
					{Name: "service-instance_bananas"},
					{Name: "service-instance_trees"},
				}
				director.VerifyAndMock(
					mockbosh.Deployments().RespondsWithJson(expectedDeployments),
				)
			})

			It("returns the deployments", func() {
				Expect(actualDeployments).To(Equal(expectedDeployments))
			})

			It("does not error", func() {
				Expect(actualDeploymentsError).NotTo(HaveOccurred())
			})
		})

		Context("when bosh returns a client error (HTTP 404)", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deployments().NotFound(),
				)
			})

			It("wraps the error", func() {
				Expect(actualDeploymentsError).To(MatchError(ContainSubstring("expected status 200, was 404")))
			})
		})

		Context("when bosh fails to fetch the task", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deployments().Fails("because reasons"),
				)
			})

			It("wraps the error", func() {
				Expect(actualDeploymentsError).To(MatchError(ContainSubstring("expected status 200, was 500")))
			})
		})
	})

	Context("deserialization", func() {
		It("unmarshals deployments response", func() {
			data := []byte(`[
			  {
			    "name": "service-instance_monkeys",
			    "cloud_config": "latest",
			    "releases": [
			      {
			        "name": "monkeys",
			        "version": "42"
			      }
			    ],
			    "stemcells": [
			      {
			        "name": "bosh-warden-boshlite-ubuntu-trusty-go_agent",
			        "version": "3312.7"
			      }
			    ]
			  },
			  {
			    "name": "service-instance_bananas",
			    "cloud_config": "latest",
			    "releases": [
			      {
			        "name": "bananas",
			        "version": "101"
			      }
			    ],
			    "stemcells": [
			      {
			        "name": "bosh-warden-boshlite-ubuntu-trusty-go_agent",
			        "version": "3312.7"
			      }
			    ]
			  }
			]`)
			var deployments []boshclient.BoshDeployment
			Expect(json.Unmarshal(data, &deployments)).To(Succeed())

			Expect(deployments).To(Equal([]boshclient.BoshDeployment{
				{Name: "service-instance_monkeys"},
				{Name: "service-instance_bananas"},
			}))
		})
	})
})
