package boshclient_test

import (
	"errors"

	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("getting manifests", func() {
	Describe("downloading raw deployment manifest", func() {
		var (
			deploymentName   string = "some-deployment"
			manifest         []byte
			rawManifest      = []byte("a-raw-manifest")
			manifestFetchErr error
		)

		Context("when the bosh director can be reached", func() {
			JustBeforeEach(func() {
				manifest, manifestFetchErr = c.GetManifestRaw(deploymentName, logger)
			})

			Context("when the deployment exists", func() {
				BeforeEach(func() {
					director.VerifyAndMock(
						mockbosh.Manifest(deploymentName).RespondsWith(rawManifest),
					)
				})

				It("calls the authorization header builder", func() {
					Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
				})

				It("returns the manifest", func() {
					Expect(manifest).To(Equal(rawManifest))
				})

				It("returns no error", func() {
					Expect(manifestFetchErr).NotTo(HaveOccurred())
				})
			})

			Context("when the deployment does not exist", func() {
				BeforeEach(func() {
					director.VerifyAndMock(
						mockbosh.Manifest(deploymentName).NotFound(),
					)
				})

				It("returns the manifest", func() {
					Expect(manifest).To(BeNil())
				})

				It("returns no error", func() {
					Expect(manifestFetchErr).NotTo(HaveOccurred())
				})
			})

			Context("when the Authorization header cannot be generated", func() {
				BeforeEach(func() {
					authHeaderBuilder.BuildReturns("", errors.New("some-error"))
				})

				It("returns an error", func() {
					Expect(manifestFetchErr).To(MatchError(ContainSubstring("some-error")))
				})
			})
		})

		Context("when the BOSH director cannot be reached", func() {
			It("returns a bosh request error", func() {
				c, err := boshclient.New("http://localhost", authorizationheader.NewBasicAuthHeaderBuilder("", ""), false, nil)
				Expect(err).NotTo(HaveOccurred())

				_, manifestFetchErr = c.GetManifestRaw(deploymentName, logger)

				Expect(manifestFetchErr).To(MatchError(ContainSubstring("error reaching bosh director")))
				Expect(manifestFetchErr).To(BeAssignableToTypeOf(boshclient.RequestError{}))
			})
		})
	})

	Describe("downloading and parsing the deployment manifest", func() {
		var (
			deploymentName   string = "some-deployment"
			manifest         *bosh.BoshManifest
			manifestFetchErr error
		)

		JustBeforeEach(func() {
			manifest, manifestFetchErr = c.GetManifest(deploymentName, logger)
		})

		Context("when the deployment exists", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Manifest(deploymentName).RespondsWith([]byte(`name: some-deployment
releases: []
stemcells: []
instance_groups: []
update:
canaries: 0
canary_watch_time: ""
update_watch_time: ""
max_in_flight: 0
`)),
				)
			})

			It("calls the authorization header builder", func() {
				Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
			})

			It("returns the manifest", func() {
				Expect(manifest.Name).To(Equal(deploymentName))
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})

			Context("when the BOSH director is serving HTTPS using a self-signed certificate", func() {
				BeforeEach(func() {
					director.Close()
					director = mockbosh.NewTLS()
					director.ExpectedAuthorizationHeader(expectedAuthHeader)
					director.VerifyAndMock(
						mockbosh.Manifest(deploymentName).RespondsWith([]byte("name: some-deployment")),
					)
				})

				It("returns no error", func() {
					Expect(manifestFetchErr).NotTo(HaveOccurred())
				})
			})
		})

		Context("when the Authorization header cannot be generated", func() {
			BeforeEach(func() {
				authHeaderBuilder.BuildReturns("", errors.New("some-error"))
			})

			It("returns an error", func() {
				Expect(manifestFetchErr).To(MatchError(ContainSubstring("some-error")))
			})
		})

		Context("when the deployment does not exist", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Manifest(deploymentName).NotFound(),
				)
			})

			It("returns the manifest", func() {
				Expect(manifest).To(BeNil())
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the manifest cannot be fetched from BOSH", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Manifest(deploymentName).Fails("because reasons"),
				)
			})

			It("returns an error", func() {
				Expect(manifestFetchErr).To(MatchError(ContainSubstring("expected status 200")))
			})
		})
	})
})
