package broker

import (
	"context"
	"errors"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

func (b *Broker) Bind(_ context.Context, instanceID, bindingID string, details brokerapi.BindDetails) (brokerapi.Binding, error) {
	logger := b.loggerFactory.NewLogger()

	errs := func(err error) (brokerapi.Binding, error) {
		return brokerapi.Binding{}, convertToDisplayableError(err, logger)
	}

	vms, manifest, err := b.getDeploymentInfo(instanceID, logger)
	switch err.(type) {
	case boshclient.DeploymentNotFoundError:
		return errs(NewDisplayableError(brokerapi.ErrInstanceDoesNotExist, errors.New("error binding: instance "+instanceID+", not found")))
	case error:
		return errs(fmt.Errorf("gathering binding info %s", err))
	}

	logger.Printf("service adapter will create binding with ID %s for instance %s\n", bindingID, instanceID)
	mappedParams, err := convertToMap(details)
	if err != nil {
		return errs(fmt.Errorf("converting to map %s", err))
	}

	adapterBinding, err := b.serviceAdapter.CreateBinding(bindingID, vms, manifest, mappedParams, logger)

	switch err := err.(type) {
	case serviceadapter.BindingAlreadyExistsError:
		return errs(NewDisplayableError(brokerapi.ErrBindingAlreadyExists, fmt.Errorf("creating binding: %s", err)))
	case serviceadapter.AppGuidNotProvidedError:
		return errs(NewDisplayableError(brokerapi.ErrAppGuidNotProvided, fmt.Errorf("creating binding: %s", err)))
	case error:
		return errs(err)
	}

	if b.credentialStore.Enabled {
		credentialId := fmt.Sprintf("%s/%s", instanceID, bindingID)
		err := b.credentialStore.PutCredentials(credentialId, adapterBinding.Credentials)
		if err != nil {
			logger.Printf("Unable to put %s in credhub: %s\n", credentialId, err)
		}
	}

	return brokerapi.Binding{
		Credentials:     adapterBinding.Credentials,
		SyslogDrainURL:  adapterBinding.SyslogDrainURL,
		RouteServiceURL: adapterBinding.RouteServiceURL,
	}, nil
}
