package broker_test

import (
	"context"
	"errors"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Bind", func() {
	var (
		instanceID             = "a-very-impressive-instance"
		bindingID              = "I'm a binding"
		serviceDeploymentName  = deploymentName(instanceID)
		adapterBindingResponse = serviceadapter.Binding{
			Credentials:     map[string]interface{}{"foo": "bar"},
			RouteServiceURL: "route",
			SyslogDrainURL:  "syslog",
		}
		boshVms             = bosh.BoshVMs{"redis-server": []string{"an.ip"}}
		actualManifest      = []byte("valid manifest")
		arbitraryParameters = map[string]interface{}{"burns": "snrub"}

		bindResult brokerapi.Binding
		bindErr    error

		bindRequest brokerapi.BindDetails
	)

	BeforeEach(func() {
		boshClient.VMsReturns(boshVms, nil)
		boshClient.GetManifestRawReturns(actualManifest, nil)
		serviceAdapter.CreateBindingReturns(adapterBindingResponse, nil)
		bindRequest = brokerapi.BindDetails{
			AppGUID:   "app_guid",
			PlanID:    "plan_id",
			ServiceID: "service_id",
			BindResource: &brokerapi.BindResource{
				AppGuid: "app_guid",
			},
			Parameters: arbitraryParameters,
		}
	})

	JustBeforeEach(func() {
		bindResult, bindErr = b.Bind(context.Background(), instanceID, bindingID, bindRequest)
	})

	It("asks bosh for VMs from a deployment named by the manifest generator", func() {
		Expect(boshClient.VMsCallCount()).To(Equal(1))
		actualServiceDeploymentName, _ := boshClient.VMsArgsForCall(0)
		Expect(actualServiceDeploymentName).To(Equal(serviceDeploymentName))
	})

	It("creates the binding using the bosh topology and admin credentials", func() {
		Expect(serviceAdapter.CreateBindingCallCount()).To(Equal(1))
		passedBindingID, passedVms, passedManifest, passedRequestParameters, _ := serviceAdapter.CreateBindingArgsForCall(0)
		Expect(passedBindingID).To(Equal(bindingID))
		Expect(passedVms).To(Equal(boshVms))
		Expect(passedManifest).To(Equal(actualManifest))
		Expect(passedRequestParameters).To(Equal(map[string]interface{}{
			"app_guid":   "app_guid",
			"plan_id":    "plan_id",
			"service_id": "service_id",
			"bind_resource": map[string]interface{}{
				"app_guid": "app_guid",
			},
			"parameters": arbitraryParameters,
		}))
	})

	It("returns the credentials, syslog drain url, and route service url produced by the service adapter", func() {
		Expect(bindResult).To(Equal(brokerapi.Binding{
			Credentials:     adapterBindingResponse.Credentials,
			SyslogDrainURL:  adapterBindingResponse.SyslogDrainURL,
			RouteServiceURL: adapterBindingResponse.RouteServiceURL,
		}))
	})

	It("does not error", func() {
		Expect(bindErr).NotTo(HaveOccurred())
	})

	It("logs using a request ID", func() {
		Expect(logBuffer.String()).To(MatchRegexp(fmt.Sprintf(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} service adapter will create binding with ID %s for instance %s`, bindingID, instanceID)))
	})

	It("does not call the credhubclient", func() {
		Expect(credhubClient.PutCredentialsCallCount()).To(Equal(0))
	})

	Context("when credhub credentials have been specified", func() {
		BeforeEach(func() {
			credentialStore.Enabled = true
		})

		It("does call the credhubclient", func() {
			Expect(credhubClient.PutCredentialsCallCount()).To(Equal(1))
		})

		It("passes the adapter binding response to credhub", func() {
			identifier, creds := credhubClient.PutCredentialsArgsForCall(0)
			Expect(identifier).To(Equal(fmt.Sprintf("%s/%s", instanceID, bindingID)))
			Expect(creds).To(Equal(adapterBindingResponse.Credentials))
		})
	})

	Context("when the request cannot be converted to json", func() {
		BeforeEach(func() {
			bindRequest = brokerapi.BindDetails{
				AppGUID:   "app_guid",
				PlanID:    "plan_id",
				ServiceID: "service_id",
				BindResource: &brokerapi.BindResource{
					AppGuid: "app_guid",
				},
				Parameters: map[string]interface{}{
					"will fail": func() bool { return true },
				},
			}
		})
		It("should return an error", func() {
			Expect(bindErr).To(HaveOccurred())
		})
	})

	Context("when binding to a non existent instance", func() {
		BeforeEach(func() {
			boshClient.VMsReturns(nil, boshclient.DeploymentNotFoundError{})
		})

		It("returns a standard error message", func() {
			Expect(bindErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("error binding: instance " + instanceID + ", not found"))
		})
	})

	Context("when bosh fails to get VMs", func() {
		BeforeEach(func() {
			boshClient.VMsReturns(nil, errors.New("oops"))
		})

		It("returns a generic error message", func() {
			Expect(bindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("oops"))
		})
	})

	Context("when manifest is not found", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, nil)
		})

		It("returns an error", func() {
			Expect(bindErr).To(HaveOccurred())
		})
	})

	Context("when broker fails to fetch manifest", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, fmt.Errorf("problem fetching manifest"))
		})

		It("returns an error", func() {
			Expect(bindErr).To(HaveOccurred())
		})
	})

	Context("when the service adapter fails to create the binding", func() {
		Context("with a generic error", func() {
			Context("with no message for the user", func() {
				BeforeEach(func() {
					serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, errors.New("binding fail"))
				})

				It("returns a generic error message", func() {
					Expect(bindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
				})

				It("logs the error", func() {
					Expect(logBuffer.String()).To(ContainSubstring("binding fail"))
				})
			})

			Context("with an error message for the user", func() {
				BeforeEach(func() {
					serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, broker.NewDisplayableError(
						errors.New("it failed, but all is not lost dear user"),
						errors.New("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'")),
					)
				})

				It("logs operator error", func() {
					Expect(logBuffer.String()).To(
						ContainSubstring("error: stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'"),
					)
				})

				It("returns the user error", func() {
					Expect(bindErr).To(MatchError("it failed, but all is not lost dear user"))
				})
			})
		})

		Context("with a binding already exists error", func() {
			BeforeEach(func() {
				serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, serviceadapter.NewBindingAlreadyExistsError(nil))
			})

			It("returns a binding already exists error", func() {
				Expect(bindErr).To(Equal(brokerapi.ErrBindingAlreadyExists))
			})
		})

		Context("with the app_guid not provided", func() {
			BeforeEach(func() {
				serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, serviceadapter.NewAppGuidNotProvidedError(nil))
			})

			It("returns an 'app guid not provided' error", func() {
				Expect(bindErr).To(Equal(brokerapi.ErrAppGuidNotProvided))
			})
		})
	})
})
