package broker

import (
	"log"
	"strings"
	"sync"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/loggerfactory"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

type Broker struct {
	boshClient      BoshClient
	cfClient        CloudFoundryClient
	serviceAdapter  ServiceAdapterClient
	credentialStore CredentialStore
	deployer        Deployer
	deploymentLock  *sync.Mutex

	serviceOffering ServiceOffering

	loggerFactory *loggerfactory.LoggerFactory
}

func New(boshClient BoshClient, cfClient CloudFoundryClient, credentialStore CredentialStore, serviceAdapter ServiceAdapterClient,
	deployer Deployer, serviceOffering ServiceOffering, loggerFactory *loggerfactory.LoggerFactory) (*Broker, error) {

	b := &Broker{
		boshClient:      boshClient,
		cfClient:        cfClient,
		credentialStore: credentialStore,
		serviceAdapter:  serviceAdapter,
		deployer:        deployer,
		deploymentLock:  &sync.Mutex{},

		serviceOffering: serviceOffering,

		loggerFactory: loggerFactory,
	}

	if err := b.startupChecks(); err != nil {
		return nil, err
	}

	return b, nil
}

const (
	OperationTypeCreate  = OperationType("create")
	OperationTypeUpdate  = OperationType("update")
	OperationTypeUpgrade = OperationType("upgrade")
	OperationTypeDelete  = OperationType("delete")
)

type OperationType string

type OperationData struct {
	BoshTaskID    int
	OperationType OperationType
}

type InstanceState struct {
	PlanID              string
	OperationInProgress bool
}

const InstancePrefix = "service-instance_"

func deploymentName(instanceID string) string {
	return InstancePrefix + instanceID
}

func instanceID(deploymentName string) string {
	return strings.TrimPrefix(deploymentName, InstancePrefix)
}

//go:generate counterfeiter -o fakes/fake_deployer.go . Deployer
type Deployer interface {
	Deploy(instanceID, planID, deploymentName string, operationType OperationType, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string, logger *log.Logger) (int, []byte, error)
}

//go:generate counterfeiter -o fakes/fake_service_adapter_client.go . ServiceAdapterClient
type ServiceAdapterClient interface {
	GenerateManifest(serviceReleases serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan, logger *log.Logger) ([]byte, error)
	CreateBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}, logger *log.Logger) (serviceadapter.Binding, error)
	DeleteBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}, logger *log.Logger) error
	GenerateDashboardUrl(instanceID string, plan serviceadapter.Plan, manifest []byte, logger *log.Logger) (string, error)
}

//go:generate counterfeiter -o fakes/fake_bosh_client.go . BoshClient
type BoshClient interface {
	DeployRaw(manifest []byte, logger *log.Logger) (int, error)
	GetTask(taskID int, logger *log.Logger) (boshclient.BoshTask, error)
	GetTasks(deploymentName string, logger *log.Logger) ([]boshclient.BoshTask, error)
	VMs(deploymentName string, logger *log.Logger) (bosh.BoshVMs, error)
	GetDeployments(logger *log.Logger) ([]boshclient.BoshDeployment, error)
	DeleteDeployment(name string, logger *log.Logger) (int, error)
	GetManifestRaw(name string, logger *log.Logger) ([]byte, error)
	GetDirectorVersion(logger *log.Logger) (string, error)
}

//go:generate counterfeiter -o fakes/fake_cloud_foundry_client.go . CloudFoundryClient
type CloudFoundryClient interface {
	GetAPIVersion(logger *log.Logger) (string, error)
	CountInstancesOfPlan(serviceOfferingID, planID string, logger *log.Logger) (int, error)
	CountInstancesOfPlans(serviceOfferingID string, logger *log.Logger) (instanceCountByPlanID map[string]int, err error)
	GetInstanceState(serviceInstanceGUID string, logger *log.Logger) (InstanceState, error)
	GetInstancesOfServiceOffering(serviceOfferingID string, logger *log.Logger) ([]string, error)
}

//go:generate counterfeiter -o fakes/fake_credhub_client.go . CredhubClient
type CredhubClient interface {
	PutCredentials(identifier string, credentialsMap map[string]interface{}) error
}
