package broker

import (
	"encoding/json"
	"fmt"
	"log"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

func (b *Broker) getDeploymentInfo(instanceID string, logger *log.Logger) (bosh.BoshVMs, []byte, error) {
	vms, err := b.boshClient.VMs(deploymentName(instanceID), logger)
	if err != nil {
		return nil, nil, err
	}
	manifest, err := b.boshClient.GetManifestRaw(deploymentName(instanceID), logger)
	if manifest == nil {
		return nil, nil, fmt.Errorf("manifest not found for deployment: %s", instanceID)
	}

	return vms, manifest, err
}

func convertProvisionDetailsToMap(details brokerapi.ProvisionDetails) (map[string]interface{}, error) {
	requestParams, err := convertToMap(details)
	if err != nil {
		return nil, err
	}

	if len(details.RawParameters) > 0 {
		arbitraryParams := map[string]interface{}{}
		if err := json.Unmarshal(details.RawParameters, &arbitraryParams); err != nil {
			return nil, brokerapi.ErrRawParamsInvalid
		}
		requestParams["parameters"] = arbitraryParams
	}

	return requestParams, nil
}

func convertToMap(object interface{}) (map[string]interface{}, error) {
	genericMap := map[string]interface{}{}
	data, err := json.Marshal(object)
	if err != nil {
		return nil, err
	}
	err = json.Unmarshal(data, &genericMap)

	if err != nil {
		return nil, err
	}
	return genericMap, nil
}

func convertToDisplayableError(err error, logger *log.Logger) error {
	displayableErr, ok := err.(DisplayableError)
	if !ok {
		displayableErr = NewGenericError(err)
	}

	logger.Println(displayableErr)
	return displayableErr.ErrorForCFUser()
}

func serviceAdapterPlan(plan Plan, globalProperties serviceadapter.Properties) serviceadapter.Plan {
	abridgedPlan := serviceadapter.Plan{
		Properties:     mergeProperties(plan.Properties, globalProperties),
		InstanceGroups: plan.InstanceGroups,
		Update:         plan.Update,
	}
	return abridgedPlan
}

func mergeProperties(planProperties, globalProperties serviceadapter.Properties) serviceadapter.Properties {
	properties := serviceadapter.Properties{}
	// avoid modifying global properties
	for k, v := range globalProperties {
		properties[k] = v
	}
	for k, v := range planProperties {
		properties[k] = v
	}
	return properties
}
