package broker_test

import (
	"bytes"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/broker/fakes"
	"github.com/pivotal-cf/on-demand-service-broker/loggerfactory"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"

	"testing"
)

const (
	serviceOfferingID                   = "service-id"
	existingPlanID                      = "some-plan-id"
	existingPlanName                    = "I'm a plan"
	existingPlanInstanceGroupName       = "instance-group-name"
	secondExistingPlanInstanceGroupName = "instance-group-name-the-second"
	serviceInstanceLimit                = 3

	secondPlanID = "another-plan"
)

var (
	b                 *broker.Broker
	brokerCreationErr error
	boshClient        *fakes.FakeBoshClient
	cfClient          *fakes.FakeCloudFoundryClient
	credentialStore   broker.CredentialStore
	credhubClient     *fakes.FakeCredhubClient
	serviceAdapter    *fakes.FakeServiceAdapterClient
	fakeDeployer      *fakes.FakeDeployer
	serviceCatalog    broker.ServiceOffering
	logBuffer         *bytes.Buffer
	loggerFactory     *loggerfactory.LoggerFactory

	existingPlan = broker.Plan{
		ID:   existingPlanID,
		Name: existingPlanName,
		Update: &serviceadapter.Update{
			Canaries:        1,
			CanaryWatchTime: "100-200",
			UpdateWatchTime: "100-200",
			MaxInFlight:     5,
		},
		Quotas: &broker.Quotas{
			ServiceInstanceLimit: serviceInstanceLimit,
		},
		Properties: serviceadapter.Properties{
			"super": "no",
		},
		InstanceGroups: []serviceadapter.InstanceGroup{
			{
				Name:               existingPlanInstanceGroupName,
				VMType:             "vm-type",
				PersistentDiskType: "disk-type",
				Instances:          42,
				Networks:           []string{"networks"},
				AZs:                []string{"my-az1", "my-az2"},
			},
			{
				Name:      secondExistingPlanInstanceGroupName,
				VMType:    "vm-type",
				Instances: 55,
				Networks:  []string{"networks2"},
			},
		},
	}

	secondPlan broker.Plan
)

var _ = BeforeEach(func() {
	secondPlan = broker.Plan{
		ID: secondPlanID,
		Properties: serviceadapter.Properties{
			"super":             "yes",
			"a_global_property": "overrides_global_value",
		},
		InstanceGroups: []serviceadapter.InstanceGroup{
			{
				Name:               existingPlanInstanceGroupName,
				VMType:             "vm-type1",
				PersistentDiskType: "disk-type1",
				Instances:          44,
				Networks:           []string{"networks1"},
				AZs:                []string{"my-az4", "my-az5"},
			},
		},
	}

	boshClient = new(fakes.FakeBoshClient)
	boshClient.GetDirectorVersionReturns("1.3262.0", nil)
	serviceAdapter = new(fakes.FakeServiceAdapterClient)
	fakeDeployer = new(fakes.FakeDeployer)
	cfClient = new(fakes.FakeCloudFoundryClient)
	cfClient.GetAPIVersionReturns("2.57.0", nil)

	credhubClient = new(fakes.FakeCredhubClient)
	credentialStore = broker.NewCredentialStore(false, credhubClient)

	serviceCatalog = broker.ServiceOffering{
		ID:               serviceOfferingID,
		Name:             "a-cool-redis-service",
		GlobalProperties: serviceadapter.Properties{"a_global_property": "global_value"},
		Plans: []broker.Plan{
			existingPlan,
			secondPlan,
		},
	}

	logBuffer = new(bytes.Buffer)
	loggerFactory = loggerfactory.New(io.MultiWriter(GinkgoWriter, logBuffer), "[broker-unit-tests] ", log.LstdFlags)
})

var _ = JustBeforeEach(func() {

	b, brokerCreationErr = broker.New(
		boshClient,
		cfClient,
		credentialStore,
		serviceAdapter,
		fakeDeployer,
		serviceCatalog,
		loggerFactory,
	)
})

func TestBroker(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Broker Suite")
}

func deploymentName(instanceID string) string {
	return "service-instance_" + instanceID
}
