package broker

import (
	"fmt"
	"log"
	"strings"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

type deployer struct {
	boshClient      BoshClient
	serviceAdapter  ServiceAdapterClient
	serviceOffering ServiceOffering
	serviceStemcell serviceadapter.Stemcell
	serviceReleases serviceadapter.ServiceReleases
}

func NewDeployer(
	boshClient BoshClient,
	serviceAdapter ServiceAdapterClient,
	serviceOffering ServiceOffering,
	serviceStemcell serviceadapter.Stemcell,
	serviceReleases serviceadapter.ServiceReleases,
) Deployer {
	return deployer{
		boshClient:      boshClient,
		serviceAdapter:  serviceAdapter,
		serviceOffering: serviceOffering,
		serviceStemcell: serviceStemcell,
		serviceReleases: serviceReleases,
	}
}

func (d deployer) Deploy(instanceID, planID, deploymentName string, operationType OperationType, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string, logger *log.Logger) (int, []byte, error) {
	tasks, err := d.boshClient.GetTasks(deploymentName, logger)
	if err != nil {
		return 0, nil, fmt.Errorf("error getting last operation for deployment %s: %s\n", deploymentName, err)
	}

	if deploymentInProgress(tasks) {
		incompleteTasks := selectIncompleteTasks(tasks)
		return 0, nil, NewOperationInProgressError(fmt.Errorf("deployment %s is still in progress: tasks %s\n", deploymentName, reduceBoshTasksToLog(incompleteTasks)))
	}

	manifest, err := d.generateManifestForDeployment(deploymentName, instanceID, planID, requestParams, oldManifest, previousPlanID, logger)
	if err != nil {
		return 0, nil, err
	}

	taskID, err := d.boshClient.DeployRaw(manifest, logger)
	if err != nil {
		return 0, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	logger.Printf("Bosh task ID for %s instance %s is %d\n", operationType, instanceID, taskID)

	return taskID, manifest, nil
}

func (d deployer) findPlans(planID string, previousPlanID *string) (serviceadapter.Plan, *serviceadapter.Plan, error) {
	plan := d.serviceOffering.FindPlanById(planID)
	if plan == nil {
		return serviceadapter.Plan{}, nil, NewDisplayableError(fmt.Errorf("plan %s does not exist", planID), fmt.Errorf("finding plan ID: %s", planID))
	}
	var previousAdapterPlan *serviceadapter.Plan
	if previousPlanID != nil {
		previousPlan := d.serviceOffering.FindPlanById(*previousPlanID)
		if previousPlan == nil {
			return serviceadapter.Plan{}, nil, NewDisplayableError(
				fmt.Errorf("previous plan %s does not exist", *previousPlanID),
				fmt.Errorf("finding previous plan ID: %s", *previousPlanID),
			)
		}
		convertedPlan := serviceAdapterPlan(*previousPlan, d.serviceOffering.GlobalProperties)
		previousAdapterPlan = &convertedPlan
	}

	return serviceAdapterPlan(*plan, d.serviceOffering.GlobalProperties), previousAdapterPlan, nil
}

func deploymentInProgress(tasks []boshclient.BoshTask) bool {
	for _, task := range tasks {
		if task.State == boshclient.BoshTaskQueued || task.State == boshclient.BoshTaskProcessing {
			return true
		}
	}

	return false
}

func selectIncompleteTasks(tasks []boshclient.BoshTask) []boshclient.BoshTask {
	var incompleteTasks []boshclient.BoshTask
	for _, task := range tasks {
		if task.State == boshclient.BoshTaskQueued || task.State == boshclient.BoshTaskProcessing {
			incompleteTasks = append(incompleteTasks, task)
		}
	}
	return incompleteTasks
}

func (d deployer) generateManifestForDeployment(deploymentName string, instanceID string, planID string, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string, logger *log.Logger) ([]byte, error) {
	plan, previousPlan, err := d.findPlans(planID, previousPlanID)
	if err != nil {
		return nil, err
	}

	serviceDeployment := serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: d.serviceReleases, Stemcell: d.serviceStemcell}

	logger.Printf("service adapter will generate manifest for instance %s\n", instanceID)
	return d.serviceAdapter.GenerateManifest(serviceDeployment, plan, requestParams, oldManifest, previousPlan, logger)
}

func reduceBoshTasksToLog(tasks []boshclient.BoshTask) string {
	var tasksArray []string
	for _, task := range tasks {
		tasksArray = append(tasksArray, task.ToLog())
	}
	return fmt.Sprintf("[%s]", strings.Join(tasksArray, ", "))
}
