package broker_test

import (
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

var _ = Describe("Deployer", func() {
	var (
		deployer         broker.Deployer
		planID           string
		previousPlanID   *string
		requestParams    map[string]interface{}
		oldManifest      []byte
		deployedManifest []byte
		deployError      error
		operationType    broker.OperationType
		boshTaskID       = 42
		returnedTaskID   int
		instanceID       = "some-instance-id"

		manifest          = []byte("a-manifest")
		deploymentName    = "a-fancy-deployment"
		serviceDeployment serviceadapter.ServiceDeployment
	)

	BeforeEach(func() {
		serviceDeployment = serviceadapter.ServiceDeployment{
			Releases: serviceadapter.ServiceReleases{{
				Name:    "name",
				Version: "vers",
				Jobs:    []string{"a", "b"},
			}},
			Stemcell: serviceadapter.Stemcell{
				OS:      "RISC OS",
				Version: "2",
			},
		}
		deployer = broker.NewDeployer(boshClient, serviceAdapter, serviceCatalog, serviceDeployment.Stemcell, serviceDeployment.Releases)
		planID = existingPlanID
		operationType = broker.OperationTypeCreate
		previousPlanID = stringPointer(secondPlanID)
	})

	JustBeforeEach(func() {
		logger := loggerFactory.NewLogger()
		returnedTaskID, deployedManifest, deployError = deployer.Deploy(instanceID, planID, deploymentName, operationType, requestParams, oldManifest, previousPlanID, logger)
	})

	Context("deployment without previous manifest", func() {
		BeforeEach(func() {
			oldManifest = nil
			previousPlanID = nil
		})

		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				boshClient.DeployRawReturns(42, nil)
				serviceAdapter.GenerateManifestReturns(manifest, nil)
				By("not having any previous tasks")
				boshClient.GetTasksReturns([]boshclient.BoshTask{}, nil)
			})

			It("checks tasks for the deployment", func() {
				Expect(boshClient.GetTasksCallCount()).To(Equal(1))
				actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
				Expect(actualDeploymentName).To(Equal(deploymentName))
			})

			It("generates a manifest, passing service deployment information to service adapter", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				actualServiceDeployment, _, _, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualServiceDeployment).To(Equal(serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: serviceDeployment.Releases, Stemcell: serviceDeployment.Stemcell}))
			})

			It("generates a manifest, passing plan to service adapter, with merged global and plan properties", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				_, actualPlan, _, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				expectedProperties := serviceadapter.Properties{"super": "no", "a_global_property": "global_value"}
				Expect(actualPlan).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: expectedProperties, Update: existingPlan.Update}))
			})

			Context("when a plan property has the same key as a global property", func() {
				BeforeEach(func() {
					planID = secondPlanID
				})

				It("merges global and plan properties, with plan taking precedence", func() {
					Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
					_, actualPlan, _, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
					expectedProperties := serviceadapter.Properties{"super": "yes", "a_global_property": "overrides_global_value"}
					Expect(actualPlan).To(Equal(serviceadapter.Plan{InstanceGroups: secondPlan.InstanceGroups, Properties: expectedProperties, Update: secondPlan.Update}))
				})
			})

			It("generates a manifest, passing request params to service adapter", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				_, _, actualRequestParams, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualRequestParams).To(Equal(requestParams))
			})

			It("generates a manifest, passing previous manifest to service adapter", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				_, _, _, actualPreviousManifest, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualPreviousManifest).To(BeNil())
			})

			It("generates a manifest, passing previous plan to service adapter", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				_, _, _, _, actualPreviousPlan, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualPreviousPlan).To(BeNil())
			})

			It("returns the bosh task ID", func() {
				Expect(returnedTaskID).To(Equal(boshTaskID))
			})

			It("Creates a bosh deployment using generated manifest", func() {
				Expect(boshClient.DeployRawCallCount()).To(Equal(1))
				deployedManifest, _ := boshClient.DeployRawArgsForCall(0)
				Expect(deployedManifest).To(Equal(manifest))
			})

			It("return the newly generated manifest", func() {
				Expect(deployedManifest).To(Equal(manifest))
			})
		})
	})
	Context("logging", func() {
		BeforeEach(func() {
			boshClient.DeployRawReturns(42, nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
		})

		Context("for creating", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeCreate
			})

			It("logs the bosh task ID returned by the director", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for create instance %s is %d", instanceID, boshTaskID)))
			})
		})
		Context("for updating", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeUpdate
			})

			It("logs the bosh task ID returned by the director", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for update instance %s is %d", instanceID, boshTaskID)))
			})
		})
		Context("for upgrading", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeUpgrade
			})

			It("logs the bosh task ID returned by the director", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for upgrade instance %s is %d", instanceID, boshTaskID)))
			})
		})
	})

	Context("deployment with previous manifest", func() {
		BeforeEach(func() {
			oldManifest = []byte("old-manifest-fetched-from-bosh")
			boshClient.DeployRawReturns(42, nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			requestParams = map[string]interface{}{"foo": "bar"}

			delete(secondPlan.Properties, "a_global_property") // test that global properties actually get merged by removing overriding plan property in this context
		})

		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
			Expect(actualDeploymentName).To(Equal(deploymentName))
		})

		It("Creates a bosh deployment from the generated manifest", func() {
			Expect(boshClient.DeployRawCallCount()).To(Equal(1))
			deployedManifest, _ := boshClient.DeployRawArgsForCall(0)
			Expect(deployedManifest).To(Equal(manifest))
		})

		It("generates manifest with correct inputs", func() {
			Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
			actualServiceDeployment, planPassedToGenerate, actualRequestParams, actualPreviousManifest, actualPreviousPlan, _ := serviceAdapter.GenerateManifestArgsForCall(0)
			expectedPlanProperties := serviceadapter.Properties{"super": "no", "a_global_property": "global_value"}
			Expect(planPassedToGenerate).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: expectedPlanProperties, Update: existingPlan.Update}))
			Expect(actualRequestParams).To(Equal(requestParams))
			Expect(actualPreviousManifest).To(Equal(oldManifest))
			Expect(actualServiceDeployment).To(Equal(serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: serviceDeployment.Releases, Stemcell: serviceDeployment.Stemcell}))
			expectedPreviousPlanProperties := serviceadapter.Properties{"super": "yes", "a_global_property": "global_value"} // overriding plan property removed for this context
			Expect(actualPreviousPlan).To(Equal(&serviceadapter.Plan{InstanceGroups: secondPlan.InstanceGroups, Properties: expectedPreviousPlanProperties}))
		})

		It("returns the bosh task ID", func() {
			Expect(returnedTaskID).To(Equal(boshTaskID))
		})
	})
	Context("plan isnt found", func() {
		BeforeEach(func() {
			planID = "non-existant-plan"
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			requestParams = map[string]interface{}{"foo": "bar"}
		})

		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
			Expect(actualDeploymentName).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("plan non-existant-plan does not exist")))
		})
	})

	Context("previous plan isnt found", func() {
		BeforeEach(func() {
			previousPlanID = stringPointer("non-existant-plan")
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			requestParams = map[string]interface{}{"foo": "bar"}
		})

		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
			Expect(actualDeploymentName).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("previous plan non-existant-plan does not exist")))
		})
	})

	Context("when the last bosh task for deployment is queued", func() {
		var previousDoneBoshTaskID = 41
		var previousErrorBoshTaskID = 40

		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{
				{State: boshclient.BoshTaskQueued, ID: boshTaskID},
				{State: boshclient.BoshTaskDone, ID: previousDoneBoshTaskID},
				{State: boshclient.BoshTaskError, ID: previousErrorBoshTaskID},
			}, nil)
		})

		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(MatchError(ContainSubstring("{ID: %d", boshTaskID)))
			Expect(deployError).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
		})

		It("does not log the previous completed tasks for the deployment", func() {
			Expect(logBuffer.String()).NotTo(ContainSubstring("done"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("ID: %d", previousDoneBoshTaskID))
			Expect(logBuffer.String()).NotTo(ContainSubstring("error"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("ID: %d", previousErrorBoshTaskID))
		})
	})

	Context("when the last bosh task for deployment is processing", func() {
		var previousDoneBoshTaskID = 41
		var previousErrorBoshTaskID = 40

		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{
				{State: boshclient.BoshTaskProcessing, ID: boshTaskID},
				{State: boshclient.BoshTaskDone, ID: previousDoneBoshTaskID},
				{State: boshclient.BoshTaskError, ID: previousErrorBoshTaskID},
			}, nil)
		})

		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(MatchError((ContainSubstring("ID: %d", boshTaskID))))
			Expect(deployError).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
		})

		It("does not log the previous tasks for the deployment", func() {
			Expect(logBuffer.String()).NotTo(ContainSubstring("done"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("ID: %d", previousDoneBoshTaskID))
			Expect(logBuffer.String()).NotTo(ContainSubstring("error"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("ID: %d", previousErrorBoshTaskID))
		})
	})

	Context("when the last bosh task for deployment fails to fetch", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns(nil, fmt.Errorf("connection error"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("connection error")))
		})
	})

	Context("when the manifest generator fails to generate the manifest", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			serviceAdapter.GenerateManifestReturns(nil, errors.New("manifest fail"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("manifest fail")))
		})
	})

	Context("when bosh fails to deploy the release", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			boshClient.DeployRawReturns(0, errors.New("oops"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("oops")))
		})
	})
})

func stringPointer(s string) *string {
	return &s
}
