package broker

import (
	"context"
	"encoding/json"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

func (b *Broker) Deprovision(_ context.Context, instanceID string, _ brokerapi.DeprovisionDetails, asyncAllowed bool) (brokerapi.DeprovisionServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	logger := b.loggerFactory.NewLogger()

	if !asyncAllowed {
		return brokerapi.DeprovisionServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	errs := func(err error) (brokerapi.DeprovisionServiceSpec, error) {
		return brokerapi.DeprovisionServiceSpec{IsAsync: true}, convertToDisplayableError(err, logger)
	}

	logger.Printf("deleting deployment for instance %s\n", instanceID)
	taskID, err := b.boshClient.DeleteDeployment(deploymentName(instanceID), logger)
	switch err.(type) {
	case boshclient.DeploymentNotFoundError:
		return errs(NewDisplayableError(
			brokerapi.ErrInstanceDoesNotExist,
			fmt.Errorf("error deprovisioning: instance %s, not found: %s", instanceID, err),
		))
	case boshclient.RequestError:
		return errs(NewBoshRequestError("delete", err))
	case error:
		return errs(fmt.Errorf("error deprovisioning: deleting bosh deployment: %s", err))
	}

	logger.Printf("Bosh task id for Delete instance %s was %d\n", instanceID, taskID)

	operationData, err := json.Marshal(OperationData{OperationType: OperationTypeDelete, BoshTaskID: taskID})
	if err != nil {
		return errs(err)
	}

	return brokerapi.DeprovisionServiceSpec{IsAsync: true, OperationData: string(operationData)}, nil
}
