package broker

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"log"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

var descriptions = map[brokerapi.LastOperationState]map[OperationType]string{
	brokerapi.InProgress: {
		OperationTypeCreate:  "Instance provisioning in progress",
		OperationTypeUpdate:  "Instance update in progress",
		OperationTypeUpgrade: "Instance upgrade in progress",
		OperationTypeDelete:  "Instance deletion in progress",
	},
	brokerapi.Succeeded: {
		OperationTypeCreate:  "Instance provisioning completed",
		OperationTypeUpdate:  "Instance update completed",
		OperationTypeUpgrade: "Instance upgrade completed",
		OperationTypeDelete:  "Instance deletion completed",
	},
	brokerapi.Failed: {
		OperationTypeCreate:  "Instance provisioning failed",
		OperationTypeUpdate:  "Instance update failed",
		OperationTypeUpgrade: "Failed for bosh task",
		OperationTypeDelete:  "Instance deletion failed",
	},
}

func (b *Broker) LastOperation(_ context.Context, instanceID, operationData string) (brokerapi.LastOperation, error) {
	logger := b.loggerFactory.NewLogger()

	errs := func(err error) (brokerapi.LastOperation, error) {
		return brokerapi.LastOperation{}, convertToDisplayableError(err, logger)
	}

	if operationData == "" {
		err := errors.New("Request missing operation data, please check your Cloud Foundry version is v238+")
		return errs(err)
	}

	var taskAndType OperationData
	if err := json.Unmarshal([]byte(operationData), &taskAndType); err != nil {
		return errs(fmt.Errorf("operation data cannot be parsed: %s", err))
	}

	if taskAndType.BoshTaskID == 0 {
		return errs(errors.New("no task ID found in operation data"))
	}

	lastBoshTask, err := b.boshClient.GetTask(taskAndType.BoshTaskID, logger)
	if err != nil {
		return errs(fmt.Errorf("error retrieving tasks from bosh, for deployment '%s': %s", deploymentName(instanceID), err))
	}
	lastOperation := constructLastOperation(lastBoshTask, taskAndType, logger)
	logLastOperation(instanceID, lastBoshTask, taskAndType, logger)

	return lastOperation, nil
}

func constructLastOperation(boshTask boshclient.BoshTask, operationData OperationData, logger *log.Logger) brokerapi.LastOperation {
	taskState := lastOperationStateForTask(boshTask.State, logger)
	description := descriptionForOperationTask(taskState, operationData, boshTask.ID)

	return brokerapi.LastOperation{State: taskState, Description: description}
}

func lastOperationStateForTask(boshTask string, logger *log.Logger) brokerapi.LastOperationState {
	switch boshTask {
	case boshclient.BoshTaskQueued, boshclient.BoshTaskProcessing:
		return brokerapi.InProgress
	case boshclient.BoshTaskDone:
		return brokerapi.Succeeded
	case boshclient.BoshTaskError, boshclient.BoshTaskCancelled, boshclient.BoshTaskCancelling, boshclient.BoshTaskTimeout:
		return brokerapi.Failed
	default:
		logger.Printf("Unrecognised BOSH task state: %s", boshTask)
		return brokerapi.Failed
	}
}

func descriptionForOperationTask(taskState brokerapi.LastOperationState, operationData OperationData, taskID int) string {
	description := descriptions[taskState][operationData.OperationType]

	if taskState == brokerapi.Failed {
		if operationData.OperationType == OperationTypeUpgrade {
			description = fmt.Sprintf(description+": %d", taskID) // Allows upgrader to log BOSH task ID when an upgrade fails
		} else {
			description = fmt.Sprintf(description+": %s", genericErrorMessage)
		}
	}

	return description
}

func logLastOperation(instanceID string, boshTask boshclient.BoshTask, operationData OperationData, logger *log.Logger) {
	logger.Printf(
		"BOSH task ID %d status: %s %s deployment for instance %s: Description: %s Result: %s\n",
		boshTask.ID,
		boshTask.State,
		operationData.OperationType,
		instanceID,
		boshTask.Description,
		boshTask.Result,
	)
}
