package broker_test

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("LastOperation", func() {
	Context("failures", func() {
		var (
			instanceID    = "a-useful-instance"
			operationData string

			lastOpErr error
		)

		JustBeforeEach(func() {
			_, lastOpErr = b.LastOperation(context.Background(), instanceID, operationData)
		})

		Context("when task cannot be retrieved from BOSH", func() {
			BeforeEach(func() {
				operationData = `{"BoshTaskID": 42}`
				boshClient.GetTaskReturns(boshclient.BoshTask{}, errors.New("something went wrong!"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error retrieving tasks from bosh, for deployment 'service-instance_a-useful-instance': something went wrong!"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when there is no operation data present in the request", func() {
			BeforeEach(func() {
				operationData = ""
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("Request missing operation data, please check your Cloud Foundry version is v238+"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when there is no task ID in the operation data", func() {
			BeforeEach(func() {
				operationData = "{}"
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("no task ID found in operation data"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when operation data cannot be parsed", func() {
			BeforeEach(func() {
				operationData = "I'm not JSON"
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("operation data cannot be parsed"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})
	})

	Context("when the task can be retrieved", func() {
		var (
			instanceID    = "not-relevent"
			operationData []byte
			taskID        = 199
		)

		type testCase struct {
			ActualBoshTask      boshclient.BoshTask
			ActualOperationType broker.OperationType
			LogContains         string

			ExpectedLastOperation brokerapi.LastOperation
		}

		testLogMessage := func(testCase testCase) func() {
			return func() {
				expectedLogMessage := fmt.Sprintf(
					"BOSH task ID %d status: %s %s deployment for instance %s: Description: %s Result: %s\n",
					taskID,
					testCase.ActualBoshTask.State,
					testCase.ActualOperationType,
					instanceID,
					testCase.ActualBoshTask.Description,
					testCase.ActualBoshTask.Result,
				)
				Expect(logBuffer.String()).To(ContainSubstring(expectedLogMessage))
			}
		}

		testLastOperation := func(testCase testCase) func() {
			return func() {
				var (
					actualLastOperation      brokerapi.LastOperation
					actualLastOperationError error
				)

				JustBeforeEach(func() {
					var err error
					operationData, err = json.Marshal(broker.OperationData{OperationType: testCase.ActualOperationType, BoshTaskID: taskID})
					Expect(err).NotTo(HaveOccurred())

					boshClient.GetTaskReturns(testCase.ActualBoshTask, nil)
					actualLastOperation, actualLastOperationError = b.LastOperation(context.Background(), instanceID, string(operationData))
				})

				It("retrieves the task by ID", func() {
					Expect(boshClient.GetTaskCallCount()).To(Equal(1))
					actualTaskID, _ := boshClient.GetTaskArgsForCall(0)
					Expect(actualTaskID).To(Equal(taskID))
				})

				It("does not error", func() {
					Expect(actualLastOperationError).NotTo(HaveOccurred())
				})

				It("returns the state", func() {
					Expect(actualLastOperation.State).To(Equal(testCase.ExpectedLastOperation.State))
				})

				It("returns the description", func() {
					Expect(actualLastOperation.Description).To(Equal(testCase.ExpectedLastOperation.Description))
				})

				It("logs a message", func() {
					Expect(logBuffer.String()).To(ContainSubstring(testCase.LogContains))
				})

				It(fmt.Sprintf("logs the deployment, type: %s, state: %s", testCase.ActualOperationType, testCase.ActualBoshTask.State), testLogMessage(testCase))
			}
		}

		Describe("while creating", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance provisioning in progress",
					},
				}),
			)

			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance provisioning in progress",
					},
				}),
			)

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is unrecognised",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: "not the state you were looking for", Result: "who knows", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,
					LogContains:         "who knows",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Cancelled",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelled, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Cancelling",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelling, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Timed out",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskTimeout, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: "Instance provisioning completed",
					},
				}),
			)
		})

		Describe("while deleting", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance deletion in progress",
					},
				}),
			)

			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance deletion in progress",
					},
				}),
			)

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Cancelled",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelled, Result: "not good", Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Cancelling",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelling, Result: "not good", Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Timed out",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskTimeout, Result: "not good", Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is unrecognised",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: "not the state you were looking for", Result: "who knows", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,
					LogContains:         "who knows",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: "Instance deletion completed",
					},
				}),
			)
		})

		Describe("while updating", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance update in progress",
					},
				}),
			)

			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance update in progress",
					},
				}),
			)

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Cancelled",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelled, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Cancelling",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelling, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Timed out",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskTimeout, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is unrecognised",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: "not the state you were looking for", Result: "who knows", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,
					LogContains:         "who knows",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
				}),
			)

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: "Instance update completed",
					},
				}),
			)
		})

		Describe("while upgrading", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance upgrade in progress",
					},
				}),
			)

			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: "Instance upgrade in progress",
					},
				}),
			)

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
				}),
			)

			Describe("last operation is Cancelled",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelled, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
				}),
			)

			Describe("last operation is Cancelling",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskCancelling, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
				}),
			)

			Describe("last operation is Timed out",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskTimeout, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
				}),
			)

			Describe("last operation is unrecognised",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: "not the state you were looking for", Result: "who knows", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,
					LogContains:         "who knows",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
				}),
			)

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: "Instance upgrade completed",
					},
				}),
			)
		})
	})
})
