package broker_test

import (
	"errors"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

var _ = Describe("Orphan Deployments", func() {
	var (
		logger               *log.Logger
		orphans              []string
		orphanDeploymentsErr error
	)

	BeforeEach(func() {
		logger = loggerFactory.NewLogger()
	})

	JustBeforeEach(func() {
		orphans, orphanDeploymentsErr = b.OrphanDeployments(logger)
	})

	Context("when there are no instances or deployments", func() {
		It("returns an empty list", func() {
			Expect(orphanDeploymentsErr).NotTo(HaveOccurred())
			Expect(orphans).To(BeEmpty())
		})
	})

	Context("when there is an instance with a deployment", func() {
		BeforeEach(func() {
			cfClient.GetInstancesOfServiceOfferingReturns([]string{"monkeys"}, nil)
			boshClient.GetDeploymentsReturns([]boshclient.BoshDeployment{{Name: "service-instance_monkeys"}}, nil)
		})

		It("returns an empty list", func() {
			Expect(orphanDeploymentsErr).NotTo(HaveOccurred())
			Expect(orphans).To(BeEmpty())
		})
	})

	Context("when there are no instances and one deployment", func() {
		BeforeEach(func() {
			cfClient.GetInstancesOfServiceOfferingReturns([]string{}, nil)
			boshClient.GetDeploymentsReturns([]boshclient.BoshDeployment{{Name: "service-instance_monkeys"}}, nil)
		})

		It("returns a list of one orphan deployment", func() {
			Expect(orphanDeploymentsErr).NotTo(HaveOccurred())
			Expect(orphans).To(ConsistOf("service-instance_monkeys"))
		})
	})

	Context("when there is one instance and no deployments", func() {
		BeforeEach(func() {
			cfClient.GetInstancesOfServiceOfferingReturns([]string{"monkeys"}, nil)
			boshClient.GetDeploymentsReturns([]boshclient.BoshDeployment{}, nil)
		})

		It("returns a list of one orphan deployment", func() {
			Expect(orphanDeploymentsErr).NotTo(HaveOccurred())
			Expect(orphans).To(BeEmpty())
		})
	})

	Context("when there is one orphan deployment and two non-ODB deployments", func() {
		BeforeEach(func() {
			cfClient.GetInstancesOfServiceOfferingReturns([]string{"monkeys"}, nil)
			deployments := []boshclient.BoshDeployment{
				{Name: "service-instance_monkeys"},
				{Name: "not-a-service-instance"},
				{Name: "acme-deployment"},
				{Name: "service-instance_bananas"},
			}
			boshClient.GetDeploymentsReturns(deployments, nil)
		})

		It("returns a list of one orphan deployment", func() {
			Expect(orphanDeploymentsErr).NotTo(HaveOccurred())
			Expect(orphans).To(ConsistOf("service-instance_bananas"))
		})
	})

	Context("when the getting the list of instances fails", func() {
		BeforeEach(func() {
			cfClient.GetInstancesOfServiceOfferingReturns([]string{}, errors.New("error listing instances: no cf api for you"))
		})

		It("broker logs an error", func() {
			Expect(orphanDeploymentsErr).To(HaveOccurred())
			Expect(logBuffer.String()).To(ContainSubstring("error listing instances: no cf api for you"))
		})
	})

	Context("when the getting the list of deployments fails", func() {
		BeforeEach(func() {
			cfClient.GetInstancesOfServiceOfferingReturns([]string{"monkeys"}, nil)
			boshClient.GetDeploymentsReturns([]boshclient.BoshDeployment{}, errors.New("error getting deployments: cuz bosh done bad"))
		})

		It("broker logs an error", func() {
			Expect(orphanDeploymentsErr).To(HaveOccurred())
			Expect(logBuffer.String()).To(ContainSubstring("error getting deployments: cuz bosh done bad"))
		})
	})
})
