package broker_test

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

var _ = Describe("provisioning", func() {
	var (
		planID string

		serviceSpec  brokerapi.ProvisionedServiceSpec
		provisionErr error

		serviceID        = "a-service-id"
		organizationGUID = "a-cf-org"
		spaceGUID        = "a-cf-space"
		jsonParams       []byte
		arbParams        = map[string]interface{}{"foo": "bar"}

		asyncAllowed = true
	)

	BeforeEach(func() {
		planID = existingPlanID
		asyncAllowed = true
		var err error
		jsonParams, err = json.Marshal(arbParams)
		Expect(err).NotTo(HaveOccurred())
		boshClient.GetManifestRawReturns(nil, nil)
	})

	JustBeforeEach(func() {
		serviceSpec, provisionErr = b.Provision(
			context.Background(),
			"some-instance",
			brokerapi.ProvisionDetails{
				PlanID:           planID,
				RawParameters:    jsonParams,
				OrganizationGUID: organizationGUID,
				SpaceGUID:        spaceGUID,
				ServiceID:        serviceID,
			},
			asyncAllowed,
		)
	})

	Context("when bosh deploys the release successfully", func() {
		var (
			newlyGeneratedManifest []byte
			deployTaskID           = 666
		)

		BeforeEach(func() {
			newlyGeneratedManifest = []byte("a newly generated manifest")
			fakeDeployer.DeployReturns(deployTaskID, newlyGeneratedManifest, nil)
		})

		It("does not error", func() {
			Expect(provisionErr).NotTo(HaveOccurred())
		})

		It("reports that the provisioning was asynchronous", func() {
			Expect(serviceSpec.IsAsync).To(BeTrue())
		})

		It("invokes the deployer", func() {
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
			actualInstanceID, actualPlan, actualDeploymentName, actualOperationType, actualRequestParams, actualOldBostManifest, actualOldPlanID, _ := fakeDeployer.DeployArgsForCall(0)
			Expect(actualOldBostManifest).To(BeNil())
			Expect(actualRequestParams).To(Equal(map[string]interface{}{
				"plan_id":           planID,
				"parameters":        arbParams,
				"organization_guid": organizationGUID,
				"space_guid":        spaceGUID,
				"service_id":        serviceID,
			}))
			Expect(actualPlan).To(Equal(planID))
			Expect(actualInstanceID).To(Equal("some-instance"))
			Expect(actualDeploymentName).To(Equal(deploymentName("some-instance")))
			Expect(actualOperationType).To(Equal(broker.OperationTypeCreate))
			Expect(actualOldPlanID).To(BeNil())
		})

		It("returns the bosh task ID and operation type", func() {
			var operationData broker.OperationData
			Expect(json.Unmarshal([]byte(serviceSpec.OperationData), &operationData)).To(Succeed())
			Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: deployTaskID, OperationType: broker.OperationTypeCreate}))
		})

		It("invokes the adapter for the dashboard url, merging global and plan properties", func() {
			Expect(serviceAdapter.GenerateDashboardUrlCallCount()).To(Equal(1))
			instanceID, plan, boshManifest, _ := serviceAdapter.GenerateDashboardUrlArgsForCall(0)
			Expect(instanceID).To(Equal("some-instance"))
			expectedProperties := serviceadapter.Properties{"super": "no", "a_global_property": "global_value"}
			Expect(plan).To(Equal(serviceadapter.Plan{
				Properties:     expectedProperties,
				InstanceGroups: existingPlan.InstanceGroups,
				Update:         existingPlan.Update,
			}))
			Expect(boshManifest).To(Equal(newlyGeneratedManifest))
		})

		Context("adapter returns a dashboard url", func() {
			BeforeEach(func() {
				serviceAdapter.GenerateDashboardUrlReturns("http://google.com", nil)
			})

			It("returns the dashboard url", func() {
				Expect(serviceSpec.DashboardURL).To(Equal("http://google.com"))
			})

			It("does not error", func() {
				Expect(provisionErr).NotTo(HaveOccurred())
			})
		})

		Context("adapter has not implemented the dashboard url", func() {
			BeforeEach(func() {
				serviceAdapter.GenerateDashboardUrlReturns("", broker.NewServiceAdapterCommandNotImplementedError(fmt.Errorf("no dashboard!")))
			})

			It("returns the dashboard as blank", func() {
				Expect(serviceSpec.DashboardURL).To(BeEmpty())
			})

			It("does not error", func() {
				Expect(provisionErr).NotTo(HaveOccurred())
			})

			It("still returns the bosh task ID and operation type", func() {
				var operationData broker.OperationData
				Expect(json.Unmarshal([]byte(serviceSpec.OperationData), &operationData)).To(Succeed())
				Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: deployTaskID, OperationType: broker.OperationTypeCreate}))
			})
		})

		Context("adapter returns an error", func() {
			BeforeEach(func() {
				serviceAdapter.GenerateDashboardUrlReturns("", errors.New("fooo"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error: fooo"))
			})

			It("returns a standard error", func() {
				Expect(provisionErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})
		})

		Context("adapter returns an AdapterCommandError", func() {
			BeforeEach(func() {
				serviceAdapter.GenerateDashboardUrlReturns("",
					broker.NewDisplayableError(
						errors.New("it failed, but all is not lost dear user"),
						errors.New("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'"),
					),
				)
			})

			It("logs operator error only", func() {
				Expect(logBuffer.String()).To(ContainSubstring("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'"))
			})

			It("returns the user error", func() {
				Expect(provisionErr).To(MatchError("it failed, but all is not lost dear user"))
			})
		})
	})

	Context("when invalid json params are provided by the broker api", func() {
		BeforeEach(func() {
			jsonParams = []byte("not a valid json")
		})

		It("wraps the returns a raw params invalid error", func() {
			Expect(provisionErr).To(Equal(brokerapi.ErrRawParamsInvalid))
		})
	})

	Context("when no arbitrary params are passed by user", func() {
		BeforeEach(func() {
			jsonParams = []byte{}
		})

		It("return no error", func() {
			Expect(provisionErr).NotTo(HaveOccurred())
		})

		It("no arbitrary params are passed to the adapter", func() {
			_, _, _, _, actualRequestParams, _, _, _ := fakeDeployer.DeployArgsForCall(0)
			Expect(actualRequestParams["parameters"]).To(BeNil())
		})

		It("invokes the deployer", func() {
			// don't need to fully replicate the conditions above, we just want to ensure that deployment still happens
			// alternatively, use shared example?
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
		})
	})

	Context("when a deployment has a generic error", func() {
		BeforeEach(func() {
			fakeDeployer.DeployReturns(0, []byte{}, fmt.Errorf("fooo"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("error: fooo"))
		})

		It("returns a standard error", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})
	})

	Context("when getting the manifest has a bosh request error", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte{}, boshclient.NewRequestError(
				fmt.Errorf("network timeout"),
			))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("error: could not get manifest: network timeout"))
		})

		It("returns the try again later error for the user", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("Currently unable to create service instance, please try again later")))
		})
	})

	Context("when a deploy has a bosh request error", func() {
		BeforeEach(func() {
			fakeDeployer.DeployReturns(0, []byte{}, boshclient.NewRequestError(
				fmt.Errorf("error deploying instance: network timeout"),
			))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("error: error deploying instance: network timeout"))
		})

		It("returns the try again later error for the user", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("Currently unable to create service instance, please try again later")))
		})
	})

	Context("when a deployment has a user displayable error", func() {
		BeforeEach(func() {
			fakeDeployer.DeployReturns(0, []byte{}, broker.NewDisplayableError(fmt.Errorf("user message"), fmt.Errorf("operator message")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("error: operator message"))
		})

		It("returns the error", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("user message")))
		})
	})

	Context("when a deployment has a AdapterCommandError", func() {
		BeforeEach(func() {
			fakeDeployer.DeployReturns(0, nil,
				broker.NewDisplayableError(
					errors.New("it failed, but all is not lost dear user"),
					errors.New("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'"),
				),
			)
		})

		It("logs operator error only", func() {
			Expect(logBuffer.String()).To(ContainSubstring("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'"))
		})

		It("returns the user error", func() {
			Expect(provisionErr).To(MatchError("it failed, but all is not lost dear user"))
		})
	})

	Context("when a provision of an already provisioned instance is triggered", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte(`manifest: true`), nil)
		})

		It("returns an error", func() {
			Expect(provisionErr).To(Equal(brokerapi.ErrInstanceAlreadyExists))
		})
	})

	Context("when the async allowed flag is false", func() {
		BeforeEach(func() {
			asyncAllowed = false
		})
		It("returns a  error", func() {
			Expect(provisionErr).To(Equal(brokerapi.ErrAsyncRequired))
		})
	})

	Context("when the count by plan provisioner id fails", func() {
		BeforeEach(func() {
			cfClient.CountInstancesOfPlanReturns(0, fmt.Errorf("count fail"))
		})
		It("returns a standard error", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("count fail"))
		})

		It("makes no deployments", func() {
			Expect(fakeDeployer.DeployCallCount()).To(BeZero())
		})
	})

	Context("when the plan quota is reached", func() {
		BeforeEach(func() {
			cfClient.CountInstancesOfPlanReturns(serviceInstanceLimit, nil)
		})

		It("counts the instances for the right plan", func() {
			Expect(cfClient.CountInstancesOfPlanCallCount()).To(Equal(1))
			actualServiceID, actualPlanID, _ := cfClient.CountInstancesOfPlanArgsForCall(0)

			Expect(actualServiceID).To(Equal(serviceCatalog.ID))
			Expect(actualPlanID).To(Equal(planID))
		})

		It("returns an error", func() {
			Expect(provisionErr).To(HaveOccurred())
		})

		It("makes no deployments", func() {
			Expect(fakeDeployer.DeployCallCount()).To(BeZero())
		})
	})

	Context("when no quota is specified", func() {
		BeforeEach(func() {
			planID = secondPlanID
		})

		It("succeeds", func() {
			Expect(provisionErr).NotTo(HaveOccurred())
		})

		It("makes a deployment", func() {
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
		})

		It("doesn't count number of instaces", func() {
			Expect(cfClient.CountInstancesOfPlanCallCount()).To(Equal(0))
		})
	})

	Context("when plan id given is not configured", func() {
		BeforeEach(func() {
			planID = "non-existant-pland"
		})

		It("return an error", func() {
			Expect(provisionErr).To(HaveOccurred())
		})
	})

	Context("adapter returns an error while fetching the dashboard url", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateDashboardUrlReturns("", fmt.Errorf("no dashboard!"))
		})

		It("returns an error", func() {
			Expect(provisionErr).To(HaveOccurred())
		})
	})
})
