package broker_test

import (
	"errors"

	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

type ItRedeploysContext struct {
	instanceID            string
	expectedRequestParams map[string]interface{}
	oldPlanID             string
	newPlanID             string
	redeployErr           error
	operationType         broker.OperationType
	boshTaskID            int
}

func ItRedeploys(context *ItRedeploysContext) {
	var (
		instanceID            = context.instanceID
		serviceDeploymentName = deploymentName(instanceID)
		requestParams         = context.expectedRequestParams

		expectedPreviousManifest = []byte("old-manifest-fetched-from-bosh")
		oldPlanID                = context.oldPlanID
		newPlanID                = context.newPlanID
		operationType            = context.operationType
	)

	Context("deployment goes well", func() {
		var manifest = []byte("updated-manifest")

		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetManifestRawReturns(expectedPreviousManifest, nil)
			fakeDeployer.DeployReturns(context.boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: context.oldPlanID}, nil)
		})

		It("does not error", func() {
			Expect(context.redeployErr).NotTo(HaveOccurred())
		})

		It("fetches correct manifest from bosh", func() {
			Expect(boshClient.GetManifestRawCallCount()).To(Equal(1))
			actualServiceDeploymentName, _ := boshClient.GetManifestRawArgsForCall(0)
			Expect(actualServiceDeploymentName).To(Equal(serviceDeploymentName))
		})

		It("deploys with the new planID", func() {
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
			actualInstanceID, actualPlanID, actualDeploymentName, actualOperationType, actualRequestParams, actualOldManifest, actualPreviousPlanID, _ := fakeDeployer.DeployArgsForCall(0)
			Expect(actualInstanceID).To(Equal(instanceID))
			Expect(actualPlanID).To(Equal(newPlanID))
			Expect(actualDeploymentName).To(Equal(broker.InstancePrefix + instanceID))
			Expect(actualRequestParams).To(Equal(requestParams))
			Expect(actualOldManifest).To(Equal(expectedPreviousManifest))
			Expect(actualOperationType).To(Equal(operationType))
			oldPlanIDCopy := oldPlanID
			Expect(actualPreviousPlanID).To(Equal(&oldPlanIDCopy))
		})

	})

	Context("when bosh fails to fetch previous manifest", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, errors.New("no manifest for you"))
		})

		if context.operationType == broker.OperationTypeUpdate {
			It("returns a generic error message to the cf user", func() {
				Expect(context.redeployErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("no manifest for you"))
			})
		}
		if context.operationType == broker.OperationTypeUpgrade {
			It("returns a generic error message to the cf user", func() {
				Expect(context.redeployErr).To(MatchError(ContainSubstring("no manifest for you")))
			})
		}
	})
}
