package broker

import (
	"errors"
	"fmt"
	"log"
	"strings"

	"github.com/coreos/go-semver/semver"
)

func (b *Broker) startupChecks() error {
	startupLogger := b.loggerFactory.NewLoggerWithoutRequestID()

	if err := b.checkAPIVersions(startupLogger); err != nil {
		return err
	}

	if err := b.verifyExistingInstancePlanIDsUnchanged(startupLogger); err != nil {
		return err
	}
	return nil
}

func (b *Broker) verifyExistingInstancePlanIDsUnchanged(logger *log.Logger) error {
	plansToInstanceCounts, err := b.cfClient.CountInstancesOfPlans(b.serviceOffering.ID, logger)
	if err != nil {
		return err
	}
	for existingPlanID, existingPlanCount := range plansToInstanceCounts {
		if existingPlanCount > 0 && !b.serviceOffering.Plans.Contains(existingPlanID) {
			return fmt.Errorf("You cannot change the plan_id of a plan that has existing service instances")
		}
	}

	return nil
}

func (b *Broker) checkAPIVersions(startupLogger *log.Logger) error {
	var apiErrorMessages []string

	if err := checkCFAPIVersion(b.cfClient, startupLogger); err != nil {
		apiErrorMessages = append(apiErrorMessages, "CF API error: "+err.Error())
	}
	if err := checkBoshDirectorVersion(b.boshClient, startupLogger); err != nil {
		apiErrorMessages = append(apiErrorMessages, "BOSH Director error: "+err.Error())
	}

	if len(apiErrorMessages) > 0 {
		return errors.New(strings.Join(apiErrorMessages, " "))
	}

	return nil
}

func checkCFAPIVersion(cfClient CloudFoundryClient, logger *log.Logger) error {
	rawCFAPIVersion, err := cfClient.GetAPIVersion(logger)
	if err != nil {
		return err
	}

	return VersionValidator{
		name:           "Cloud Foundry",
		minVersionName: "CF v238+",
		minAPIVersion:  *semver.New("2.57.0"),
	}.validate(rawCFAPIVersion)
}

func checkBoshDirectorVersion(boshClient BoshClient, logger *log.Logger) error {
	rawDirectorVersion, err := boshClient.GetDirectorVersion(logger)
	if err != nil {
		return err
	}

	return VersionValidator{
		name:           "BOSH Director",
		minVersionName: "BOSH v257+",
		minAPIVersion:  *semver.New("1.3262.0"),
	}.validate(rawDirectorVersion)
}

type VersionValidator struct {
	name           string
	minVersionName string
	minAPIVersion  semver.Version
}

func (v VersionValidator) validate(rawVersion string) error {
	version, err := semver.NewVersion(rawVersion)
	if err != nil {
		return fmt.Errorf("%s API version couldn't be parsed. Expected a semver, got: %s.", v.name, rawVersion)
	}
	if version.LessThan(v.minAPIVersion) {
		return fmt.Errorf("%s API version is insufficient, ODB requires %s.", v.name, v.minVersionName)
	}
	return nil
}
