package broker_test

import (
	"context"
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

var _ = Describe("Unbind", func() {
	var (
		instanceID     = "a-most-unimpressive-instance"
		bindingID      = "I'm still a binding"
		serviceID      = "awesome-service"
		planID         = "awesome-plan"
		deploymentName = broker.InstancePrefix + instanceID
		boshVms        = bosh.BoshVMs{"redis-server": []string{"an.ip"}}
		actualManifest = []byte("a valid manifest")
		unbindErr      error
	)

	BeforeEach(func() {
		boshClient.VMsReturns(boshVms, nil)
		serviceAdapter.DeleteBindingReturns(nil)
		boshClient.GetManifestRawReturns(actualManifest, nil)
	})

	JustBeforeEach(func() {
		unbindErr = b.Unbind(context.Background(), instanceID, bindingID, brokerapi.UnbindDetails{ServiceID: serviceID, PlanID: planID})
	})

	It("asks bosh for VMs from a deployment named by the manifest generator", func() {
		Expect(boshClient.VMsCallCount()).To(Equal(1))
		actualDeploymentName, _ := boshClient.VMsArgsForCall(0)
		Expect(actualDeploymentName).To(Equal(deploymentName))
	})

	It("destroys the binding using the bosh topology and admin credentials", func() {
		Expect(serviceAdapter.DeleteBindingCallCount()).To(Equal(1))
		passedBindingID, passedVms, passedManifest, passedRequestParams, _ := serviceAdapter.DeleteBindingArgsForCall(0)
		Expect(passedBindingID).To(Equal(bindingID))
		Expect(passedVms).To(Equal(boshVms))
		Expect(passedManifest).To(Equal(actualManifest))
		Expect(passedRequestParams).To(Equal(map[string]interface{}{"service_id": serviceID, "plan_id": planID}))
	})

	It("does not error", func() {
		Expect(unbindErr).NotTo(HaveOccurred())
	})

	It("logs using a request ID", func() {
		Expect(logBuffer.String()).To(MatchRegexp(fmt.Sprintf(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} service adapter will delete binding with ID %s for instance %s`, bindingID, instanceID)))
	})

	Context("when bosh fails to get VMs", func() {
		BeforeEach(func() {
			boshClient.VMsReturns(nil, errors.New("oops"))
		})

		It("returns a generic error message to the cf user", func() {
			Expect(unbindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("oops"))
		})
	})

	Context("when bosh client does not return a manifest", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, nil)
		})

		It("returns an error", func() {
			Expect(unbindErr).To(HaveOccurred())
		})
	})

	Context("when there is an error while fetching a manifest from the bosh client", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, fmt.Errorf("problem fetching manifest"))
		})

		It("returns an error", func() {
			Expect(unbindErr).To(HaveOccurred())
		})
	})

	Context("when cannot find the instance", func() {
		BeforeEach(func() {
			boshClient.VMsReturns(bosh.BoshVMs{}, boshclient.DeploymentNotFoundError{})
		})

		It("returns an error", func() {
			Expect(unbindErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
		})
	})

	Context("when the service adapter fails to destroy the binding", func() {
		Context("with no message for the user", func() {
			BeforeEach(func() {
				serviceAdapter.DeleteBindingReturns(errors.New("executing unbinding failed"))
			})

			It("returns a generic error message to the cf user", func() {
				Expect(unbindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})

			It("logs the error for the operator", func() {
				Expect(logBuffer.String()).To(ContainSubstring("executing unbinding failed"))
			})
		})

		Context("with an error message for the user", func() {
			BeforeEach(func() {
				serviceAdapter.DeleteBindingReturns(broker.NewDisplayableError(
					errors.New("it failed, but all is not lost dear user"),
					errors.New("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'")),
				)
			})

			It("logs operator error only", func() {
				Expect(logBuffer.String()).To(
					ContainSubstring("stdout: 'it failed, but all is not lost dear user', stderr: 'something helpful for the operator'"),
				)
			})

			It("returns the user error", func() {
				Expect(unbindErr).To(MatchError("it failed, but all is not lost dear user"))
			})
		})
	})

	Context("when the service adapter cannot find the binding", func() {
		BeforeEach(func() {
			serviceAdapter.DeleteBindingReturns(serviceadapter.NewBindingNotFoundError(nil))
		})

		It("returns a binding not found error", func() {
			Expect(unbindErr).To(Equal(brokerapi.ErrBindingDoesNotExist))
		})
	})
})
