package broker

import (
	"context"
	"encoding/json"
	"fmt"
	"log"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

func (b *Broker) Update(_ context.Context, instanceID string, details brokerapi.UpdateDetails, asyncAllowed bool) (brokerapi.UpdateServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	logger := b.loggerFactory.NewLogger()

	errs := func(err error) (brokerapi.UpdateServiceSpec, error) {
		return brokerapi.UpdateServiceSpec{IsAsync: true}, convertToDisplayableError(err, logger)
	}

	if err := b.validatePlanQuota(details, logger); err != nil {
		return errs(err)
	}

	logger.Printf("updating instance %s", instanceID)
	detailsMap, err := convertToMap(details)
	if err != nil {
		return errs(err)
	}

	boshTaskID, err := b.redeploy(instanceID, details.PlanID, OperationTypeUpdate, detailsMap, &details.PreviousValues.PlanID, logger)
	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("update", fmt.Errorf("error deploying instance: %s", err)))
	case error:
		return errs(fmt.Errorf("error deploying instance: %s", err))
	}

	if err != nil {
		return errs(err)
	}

	operationData, err := json.Marshal(OperationData{OperationType: OperationTypeUpdate, BoshTaskID: boshTaskID})
	if err != nil {
		return errs(err)
	}

	return brokerapi.UpdateServiceSpec{IsAsync: true, OperationData: string(operationData)}, nil
}

func (b *Broker) validatePlanQuota(details brokerapi.UpdateDetails, logger *log.Logger) error {
	if details.PreviousValues.PlanID == details.PlanID {
		return nil
	}

	planConfig := b.serviceOffering.FindPlanById(details.PlanID)
	if planConfig == nil {
		return NewDisplayableError(fmt.Errorf("plan %s not found", details.PlanID), fmt.Errorf("finding plan ID %s", details.PlanID))
	}

	count, err := b.cfClient.CountInstancesOfPlan(details.ServiceID, details.PlanID, logger)
	if err != nil {
		return fmt.Errorf("error counting instances of plan: %s", err)
	}

	if planConfig.Quotas != nil && count >= planConfig.Quotas.ServiceInstanceLimit {
		return NewDisplayableError(brokerapi.ErrPlanQuotaExceeded, fmt.Errorf("plan quota exceeded for plan ID %s", details.PlanID))
	}

	return nil
}
