package broker_test

import (
	"context"
	"encoding/json"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Update", func() {
	var (
		updateSpec      brokerapi.UpdateServiceSpec
		updateDetails   brokerapi.UpdateDetails
		arbitraryParams = map[string]interface{}{"foo": "bar"}
		serviceID       = "serviceID"
		orgGUID         = "organizationGUID"
		spaceGUID       = "spaceGUID"
		boshTaskID      = 447
	)

	Context("redeploy shared example", func() {
		var (
			testContext = ItRedeploysContext{
				instanceID: "some-instance",

				expectedRequestParams: map[string]interface{}{
					"plan_id":    existingPlanID,
					"service_id": serviceID,
					"previous_values": map[string]interface{}{
						"plan_id":         secondPlanID,
						"service_id":      serviceID,
						"organization_id": orgGUID,
						"space_id":        spaceGUID,
					},
					"parameters": arbitraryParams,
				},
				oldPlanID:     secondPlanID,
				newPlanID:     existingPlanID,
				operationType: broker.OperationTypeUpdate,
				boshTaskID:    512,
			}
			updateDetails = brokerapi.UpdateDetails{
				PlanID:     testContext.newPlanID,
				Parameters: arbitraryParams,
				ServiceID:  serviceID,
				PreviousValues: brokerapi.PreviousValues{
					PlanID:    secondPlanID,
					OrgID:     orgGUID,
					ServiceID: serviceID,
					SpaceID:   spaceGUID,
				},
			}
		)

		JustBeforeEach(func() {
			updateSpec, testContext.redeployErr = b.Update(context.Background(), testContext.instanceID, updateDetails, true)
		})

		ItRedeploys(&testContext)
	})

	Context("when there is a previous deployment for the service instance", func() {
		var (
			updateError error
			newPlanID   string
			oldPlanID   string
			instanceID  = "some-instance-id"
		)

		JustBeforeEach(func() {
			updateDetails = brokerapi.UpdateDetails{
				PlanID:     newPlanID,
				Parameters: arbitraryParams,
				ServiceID:  serviceID,
				PreviousValues: brokerapi.PreviousValues{
					PlanID:    oldPlanID,
					OrgID:     orgGUID,
					ServiceID: serviceID,
					SpaceID:   spaceGUID,
				},
			}
			updateSpec, updateError = b.Update(context.Background(), instanceID, updateDetails, true)
		})

		Context("when changing plan", func() {
			BeforeEach(func() {
				newPlanID = existingPlanID
				oldPlanID = secondPlanID
			})

			Context("when the new plan's quota has not been met", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlanReturns(0, nil)
					serviceAdapter.GenerateManifestReturns([]byte("an-updated-manifest"), nil)
					boshClient.GetManifestRawReturns([]byte("some-old-manifest"), nil)
					fakeDeployer.DeployReturns(boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
				})

				It("counts the instances of the plan in Cloud Controller", func() {
					Expect(cfClient.CountInstancesOfPlanCallCount()).To(Equal(1))
					actualServiceOfferingID, actualPlanID, _ := cfClient.CountInstancesOfPlanArgsForCall(0)
					Expect(actualServiceOfferingID).To(Equal(serviceID))
					Expect(actualPlanID).To(Equal(newPlanID))
				})

				It("returns that is operated asynchronously", func() {
					Expect(updateSpec.IsAsync).To(BeTrue())
				})

				It("returns the bosh task ID and operation type", func() {
					Expect(updateError).NotTo(HaveOccurred())
					var operationData broker.OperationData
					Expect(json.Unmarshal([]byte(updateSpec.OperationData), &operationData)).To(Succeed())
					Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: boshTaskID, OperationType: broker.OperationTypeUpdate}))
				})

				It("logs with a request ID", func() {
					Expect(logBuffer.String()).To(MatchRegexp(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} updating instance`))
				})
			})

			Context("when the new plan's quota has been reached", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlanReturns(serviceInstanceLimit, nil)
				})

				It("returns an error", func() {
					Expect(updateError).To(HaveOccurred())
				})

				It("does not redeploy", func() {
					Expect(serviceAdapter.GenerateManifestCallCount()).To(BeZero())
					Expect(boshClient.GetManifestRawCallCount()).To(BeZero())
					Expect(fakeDeployer.DeployCallCount()).To(BeZero())
				})
			})
		})

		Context("when the plan is not changed", func() {
			BeforeEach(func() {
				newPlanID = existingPlanID
				oldPlanID = existingPlanID
				serviceAdapter.GenerateManifestReturns([]byte("an-updated-manifest"), nil)
				boshClient.GetManifestRawReturns([]byte("some-old-manifest"), nil)
				fakeDeployer.DeployReturns(boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			})

			Context("when the new plan's quota has not been met", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlanReturns(0, nil)
				})

				It("returns that is operated asynchronously", func() {
					Expect(updateSpec.IsAsync).To(BeTrue())
				})

				It("returns the bosh task ID and operation type", func() {
					Expect(updateError).NotTo(HaveOccurred())
					var operationData broker.OperationData
					Expect(json.Unmarshal([]byte(updateSpec.OperationData), &operationData)).To(Succeed())
					Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: boshTaskID, OperationType: broker.OperationTypeUpdate}))
				})
			})

			Context("when the plan's quota has been met", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlanReturns(serviceInstanceLimit, nil)
				})

				It("returns that is operated asynchronously", func() {
					Expect(updateSpec.IsAsync).To(BeTrue())
				})

				It("returns the bosh task ID and operation type", func() {
					Expect(updateError).NotTo(HaveOccurred())
					var operationData broker.OperationData
					Expect(json.Unmarshal([]byte(updateSpec.OperationData), &operationData)).To(Succeed())
					Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: boshTaskID, OperationType: broker.OperationTypeUpdate}))
				})
			})
		})

		Context("when the plan cannot be found in config", func() {
			BeforeEach(func() {
				newPlanID = "non-existant-plan-id"
				oldPlanID = secondPlanID
			})

			It("returns a user displayable error", func() {
				Expect(updateError).To(MatchError(ContainSubstring(fmt.Sprintf("plan %s not found", newPlanID))))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error: finding plan ID %s", newPlanID)))
			})

			It("does not redeploy", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(BeZero())
				Expect(boshClient.GetManifestRawCallCount()).To(BeZero())
				Expect(fakeDeployer.DeployCallCount()).To(BeZero())
			})
		})

		Context("when the service instances for the plan cannot be counted", func() {
			BeforeEach(func() {
				newPlanID = existingPlanID
				oldPlanID = secondPlanID
				cfClient.CountInstancesOfPlanReturns(0, fmt.Errorf("count fail"))
			})

			It("returns a standard error", func() {
				Expect(updateError).To(MatchError(ContainSubstring("An internal error has occurred")))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error: error counting instances of plan: count fail"))
			})

			It("does not redeploy", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(BeZero())
				Expect(boshClient.GetManifestRawCallCount()).To(BeZero())
				Expect(fakeDeployer.DeployCallCount()).To(BeZero())
			})
		})

		Context("when the bosh director is unavailable", func() {
			BeforeEach(func() {
				newPlanID = existingPlanID
				oldPlanID = existingPlanID
				cfClient.CountInstancesOfPlanReturns(0, nil)
				serviceAdapter.GenerateManifestReturns([]byte("an-updated-manifest"), nil)
			})

			Context("when getting the old manifest has a bosh request error", func() {
				BeforeEach(func() {
					boshClient.GetManifestRawReturns([]byte("some-old-manifest"), boshclient.NewRequestError(
						fmt.Errorf("network timeout"),
					))
				})

				It("logs the error", func() {
					Expect(logBuffer.String()).To(ContainSubstring("error: error deploying instance: network timeout"))
				})

				It("returns the try again later error for the user", func() {
					Expect(updateError).To(MatchError(ContainSubstring("Currently unable to update service instance, please try again later")))
				})
			})

			Context("when a deploy has a bosh request error", func() {
				BeforeEach(func() {
					boshClient.GetManifestRawReturns([]byte("some-old-manifest"), nil)
					fakeDeployer.DeployReturns(0, []byte{}, boshclient.NewRequestError(
						fmt.Errorf("network timeout"),
					))
				})

				It("logs the error", func() {
					Expect(logBuffer.String()).To(ContainSubstring("error: error deploying instance: network timeout"))
				})

				It("returns the try again later error for the user", func() {
					Expect(updateError).To(MatchError(ContainSubstring("Currently unable to update service instance, please try again later")))
				})
			})
		})
	})
})
