package broker_test

import (
	"errors"
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Upgrade", func() {
	var (
		upgradeTaskID int
		instanceID    = "some-instance"
		logger        *log.Logger

		context = ItRedeploysContext{
			instanceID:    instanceID,
			oldPlanID:     existingPlanID,
			newPlanID:     existingPlanID,
			operationType: broker.OperationTypeUpgrade,
			boshTaskID:    876,
		}
	)

	JustBeforeEach(func() {
		logger = loggerFactory.NewLogger()
		upgradeTaskID, context.redeployErr = b.Upgrade(context.instanceID, logger)
	})

	ItRedeploys(&context)

	Context("when there is a previous deployment for the service instance", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns([]byte("fancy-new-manifest"), nil)
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			fakeDeployer.DeployReturns(context.boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("returns the task ID for the upgrade task", func() {
			Expect(upgradeTaskID).To(Equal(context.boshTaskID))
		})

		It("fetches correct instance", func() {
			Expect(cfClient.GetInstanceStateCallCount()).To(Equal(1))
			actualInstanceID, _ := cfClient.GetInstanceStateArgsForCall(0)
			Expect(actualInstanceID).To(Equal(instanceID))
		})

		It("logs with a request ID", func() {
			Expect(logBuffer.String()).To(MatchRegexp(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} upgrading instance`))
		})

		Context("and there is an operation in progress on the instance", func() {
			BeforeEach(func() {
				cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID, OperationInProgress: true}, nil)
			})

			It("returns an OperationInProgressError", func() {
				Expect(context.redeployErr).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
			})
		})
	})

	Context("when the service instance does not exist in Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{}, broker.NewNotFoundError())
		})

		It("returns a DeploymentNotFoundError", func() {
			Expect(context.redeployErr).To(BeAssignableToTypeOf(broker.DeploymentNotFoundError{}))
		})
	})

	Context("when the plan details cannot be retrieved from Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{}, errors.New("no instance for you"))
		})

		It("returns the error to the cf user", func() {
			Expect(context.redeployErr).To(MatchError(ContainSubstring("no instance for you")))
		})
	})

	Context("when there is a task in progress on the instance", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns([]byte("fancy-new-manifest"), nil)
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			fakeDeployer.DeployReturns(0, []byte{}, broker.NewOperationInProgressError(fmt.Errorf("deployment %s is still in progress", deploymentName(instanceID))))
		})

		It("returns an OperationInProgressError", func() {
			Expect(context.redeployErr).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
		})
	})
})
