package cloud_foundry_client

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

type Client struct {
	httpJsonClient
	url string
}

func New(url string, authHeaderBuilder authorizationheader.AuthHeaderBuilder, trustedCertPEM []byte, disableTLSCertVerification bool) (Client, error) {
	httpClient, err := newWrappedHttpClient(authHeaderBuilder, trustedCertPEM, disableTLSCertVerification)
	if err != nil {
		return Client{}, err
	}
	return Client{httpJsonClient: httpClient, url: url}, nil
}

func (c Client) CountInstancesOfPlans(serviceID string, logger *log.Logger) (map[string]int, error) {
	plans, err := c.getPlansForServiceID(serviceID, logger)
	if err != nil {
		return map[string]int{}, err
	}

	output := map[string]int{}
	for _, plan := range plans {
		count, err := c.countServiceInstancesOfServicePlan(plan.ServicePlanEntity.ServiceInstancesUrl, logger)
		if err != nil {
			return nil, err
		}
		output[plan.ServicePlanEntity.UniqueID] = count
	}

	return output, nil
}

func (c Client) GetInstanceState(serviceInstanceGUID string, logger *log.Logger) (broker.InstanceState, error) {
	instance, err := c.getServiceInstance(serviceInstanceGUID, logger)
	if err != nil {
		return broker.InstanceState{}, err
	}

	plan, err := c.getServicePlan(instance.Entity.ServicePlanURL, logger)
	if err != nil {
		return broker.InstanceState{}, err
	}

	return broker.InstanceState{PlanID: plan.ServicePlanEntity.UniqueID, OperationInProgress: instance.Entity.LastOperation.State == "in progress"}, nil
}

func (c Client) CountInstancesOfPlan(serviceID, servicePlanID string, logger *log.Logger) (int, error) {
	plans, err := c.getPlansForServiceID(serviceID, logger)
	if err != nil {
		return 0, err
	}

	for _, plan := range plans {
		if plan.ServicePlanEntity.UniqueID == servicePlanID {
			count, err := c.countServiceInstancesOfServicePlan(plan.ServicePlanEntity.ServiceInstancesUrl, logger)
			if err != nil {
				return 0, err
			}
			return count, nil
		}
	}

	return 0, fmt.Errorf("service plan %s not found for service %s", servicePlanID, serviceID)
}

func (c Client) GetInstancesOfServiceOffering(serviceOfferingID string, logger *log.Logger) ([]string, error) {
	plans, err := c.getPlansForServiceID(serviceOfferingID, logger)
	if err != nil {
		return nil, err
	}

	var instances []string
	for _, plan := range plans {
		var serviceInstancesResp serviceInstancesResponse
		if err := c.Get(fmt.Sprintf("%s/v2/service_plans/%s/service_instances", c.url, plan.Metadata.GUID), &serviceInstancesResp, logger); err != nil {
			return nil, err
		}
		for _, instance := range serviceInstancesResp.ServiceInstances {
			instances = append(instances, instance.Metadata.GUID)
		}
	}
	return instances, nil
}

func (c Client) GetAPIVersion(logger *log.Logger) (string, error) {
	var infoResponse infoResponse
	err := c.Get(fmt.Sprintf("%s/v2/info", c.url), &infoResponse, logger)
	if err != nil {
		return "", err
	}
	return infoResponse.APIVersion, nil
}

func (c Client) getPlansForServiceID(serviceID string, logger *log.Logger) ([]ServicePlan, error) {
	requiredService, err := c.findServiceByUniqueID(serviceID, logger)
	if err != nil {
		return nil, err
	}

	if requiredService == nil {
		return nil, nil
	}

	return c.listAllPlans(requiredService.ServiceEntity.ServicePlansUrl, logger)
}

const DefaultPerPage = "100"

func (c Client) listServices(path string, logger *log.Logger) (serviceResponse, error) {
	resp := serviceResponse{}
	return resp, c.Get(fmt.Sprintf("%s%s", c.url, path), &resp, logger)
}

func (c Client) findServiceByUniqueID(uniqueID string, logger *log.Logger) (*service, error) {
	path := "/v2/services?results-per-page=" + DefaultPerPage
	for path != "" {
		resp, err := c.listServices(path, logger)
		if err != nil {
			return nil, err
		}
		if service := resp.Services.findByUniqueID(uniqueID); service != nil {
			return service, nil
		}
		path = resp.NextPath
	}

	return nil, nil
}

func (c Client) getServiceInstance(serviceInstanceGUID string, logger *log.Logger) (serviceInstance, error) {
	path := fmt.Sprintf("/v2/service_instances/%s", serviceInstanceGUID)
	var instance serviceInstance
	err := c.Get(fmt.Sprintf("%s%s", c.url, path), &instance, logger)
	return instance, err
}

func (c Client) getServicePlan(servicePlanPath string, logger *log.Logger) (ServicePlan, error) {
	var plan ServicePlan
	err := c.Get(fmt.Sprintf("%s%s", c.url, servicePlanPath), &plan, logger)
	return plan, err
}

func (c Client) listAllPlans(url string, logger *log.Logger) ([]ServicePlan, error) {
	plans := []ServicePlan{}
	path := url + "?results-per-page=" + DefaultPerPage
	for path != "" {
		servicePlanResponse, err := c.listPlans(path, logger)
		if err != nil {
			return nil, err
		}
		plans = append(plans, servicePlanResponse.ServicePlans...)
		path = servicePlanResponse.NextPath
	}

	return plans, nil
}
func (c Client) listPlans(path string, logger *log.Logger) (ServicePlanResponse, error) {
	servicePlanResponse := ServicePlanResponse{}
	return servicePlanResponse, c.Get(fmt.Sprintf("%s%s", c.url, path), &servicePlanResponse, logger)
}
func (c Client) countServiceInstancesOfServicePlan(path string, logger *log.Logger) (int, error) {
	resp := serviceInstancesResponse{}
	err := c.Get(fmt.Sprintf("%s%s", c.url, path), &resp, logger)
	if err != nil {
		return 0, err
	}
	return resp.TotalResults, nil
}
