package cloud_foundry_client

import (
	"crypto/x509"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/broker"

	"github.com/craigfurman/herottp"
)

type httpJsonClient struct {
	client            *herottp.Client
	AuthHeaderBuilder authorizationheader.AuthHeaderBuilder
}

func (w httpJsonClient) Get(u string, obj interface{}, logger *log.Logger) error {
	req, err := http.NewRequest(http.MethodGet, u, nil)
	if err != nil {
		return err
	}
	authHeader, err := w.AuthHeaderBuilder.Build(logger)
	if err != nil {
		return err
	}
	req.Header.Add("Authorization", authHeader)
	logger.Printf(fmt.Sprintf("GET %s", u))
	return w.readResponse(obj, req)
}

func (w httpJsonClient) PostForm(u string, data map[string]string, respObj interface{}, logger *log.Logger) error {
	form := url.Values{}
	for k, v := range data {
		form.Add(k, v)
	}
	req, err := http.NewRequest(http.MethodPost, u, strings.NewReader(form.Encode()))
	if err != nil {
		return err
	}
	authHeader, err := w.AuthHeaderBuilder.Build(logger)
	if err != nil {
		return err
	}

	req.Header.Add("Authorization", authHeader)
	req.Header.Add("Content-Type", "application/x-www-form-urlencoded")

	logger.Printf(fmt.Sprintf("POST %s", u))
	return w.readResponse(respObj, req)
}

func (w httpJsonClient) readResponse(obj interface{}, req *http.Request) error {
	response, err := w.client.Do(req)
	if err != nil {
		return err
	}
	defer response.Body.Close()
	if response.StatusCode == http.StatusNotFound {
		return broker.NewNotFoundError()
	}

	if response.StatusCode != http.StatusOK {
		body, _ := ioutil.ReadAll(response.Body)
		err := fmt.Errorf("Unexpected reponse status %d, %q", response.StatusCode, string(body))
		if response.StatusCode == http.StatusUnauthorized {
			return unauthorizedRequestError{err}
		} else {
			return err
		}
	}
	return json.NewDecoder(response.Body).Decode(&obj)
}

type unauthorizedRequestError struct {
	error
}

func newWrappedHttpClient(authHeaderBuilder authorizationheader.AuthHeaderBuilder, trustedCertPEM []byte, disableTLSCertVerification bool) (httpJsonClient, error) {
	rootCAs, err := x509.SystemCertPool()
	if err != nil {
		return httpJsonClient{}, err
	}
	rootCAs.AppendCertsFromPEM(trustedCertPEM)
	config := herottp.Config{
		DisableTLSCertificateVerification: disableTLSCertVerification,
		RootCAs: rootCAs,
		Timeout: 30 * time.Second,
	}

	return httpJsonClient{
		client:            herottp.New(config),
		AuthHeaderBuilder: authHeaderBuilder,
	}, nil
}
