package main

import (
	"flag"
	"fmt"
	"log"
	"net/http"
	"os"

	"github.com/gorilla/mux"
	apiauth "github.com/pivotal-cf/brokerapi/auth"
	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/cloud_foundry_client"
	"github.com/pivotal-cf/on-demand-service-broker/config"
	"github.com/pivotal-cf/on-demand-service-broker/credhubclient"
	"github.com/pivotal-cf/on-demand-service-broker/external"
	"github.com/pivotal-cf/on-demand-service-broker/loggerfactory"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"

	"code.cloudfoundry.org/lager"
	"github.com/pivotal-cf/brokerapi"
	"github.com/urfave/negroni"
)

func main() {
	// Yes, you really do need the space at the end of the prefix.
	prefix := "[on-demand-service-broker] "
	logFlags := log.Ldate | log.Ltime | log.Lmicroseconds | log.LUTC
	loggerFactory := loggerfactory.New(os.Stdout, prefix, logFlags)

	logger := loggerFactory.NewLoggerWithoutRequestID()
	logger.Println("Starting broker")

	configFilePath := flag.String("configFilePath", "", "path to config file")

	flag.Parse()
	if *configFilePath == "" {
		logger.Fatal("must supply -configFilePath")
	}

	conf, err := config.Parse(*configFilePath)
	if err != nil {
		logger.Fatalf("error parsing config: %s", err)
	}
	startBroker(conf, logger, loggerFactory)
}

func startBroker(conf config.Config, logger *log.Logger, loggerFactory *loggerfactory.LoggerFactory) {
	var (
		boshAuthenticator authorizationheader.AuthHeaderBuilder
		cfAuthenticator   authorizationheader.AuthHeaderBuilder
		err               error
	)

	boshAuthConfig := conf.Bosh.Authentication
	if boshAuthConfig.Basic.IsSet() {
		boshAuthenticator = authorizationheader.NewBasicAuthHeaderBuilder(
			boshAuthConfig.Basic.Username,
			boshAuthConfig.Basic.Password,
		)
	} else if boshAuthConfig.UAA.IsSet() {
		boshAuthenticator, err = authorizationheader.NewClientTokenAuthHeaderBuilder(
			boshAuthConfig.UAA.UAAURL,
			boshAuthConfig.UAA.ID,
			boshAuthConfig.UAA.Secret,
			conf.Broker.DisableSSLCertVerification,
			[]byte(conf.Bosh.TrustedCert),
		)
		if err != nil {
			logger.Fatalf("error creating authorization header builder: %s", err)
		}
	}

	boshClient, err := boshclient.New(conf.Bosh.URL, boshAuthenticator, conf.Broker.DisableSSLCertVerification, []byte(conf.Bosh.TrustedCert))
	if err != nil {
		logger.Fatalf("error creating bosh client: %s", err)
	}

	cfAuthConfig := conf.CF.Authentication
	if cfAuthConfig.ClientCredentials.IsSet() {
		cfAuthenticator, err = authorizationheader.NewClientTokenAuthHeaderBuilder(
			cfAuthConfig.URL,
			cfAuthConfig.ClientCredentials.ID,
			cfAuthConfig.ClientCredentials.Secret,
			conf.Broker.DisableSSLCertVerification,
			[]byte(conf.CF.TrustedCert),
		)
	} else {
		cfAuthenticator, err = authorizationheader.NewUserTokenAuthHeaderBuilder(
			cfAuthConfig.URL,
			"cf",
			"",
			cfAuthConfig.UserCredentials.Username,
			cfAuthConfig.UserCredentials.Password,
			conf.Broker.DisableSSLCertVerification,
			[]byte(conf.CF.TrustedCert),
		)
	}

	if err != nil {
		logger.Fatalf("error creating authorization header builder: %s", err)
	}

	cfClient, err := cloud_foundry_client.New(
		conf.CF.URL,
		cfAuthenticator,
		[]byte(conf.CF.TrustedCert),
		conf.Broker.DisableSSLCertVerification,
	)
	if err != nil {
		logger.Fatalf("error creating Cloud Foundry client: %s", err)
	}

	serviceAdapter := &external.Adapter{
		ExternalBinPath: conf.ServiceAdapter.Path,
		CommandRunner:   external.NewCommandRunner(),
	}

	deploymentManager := broker.NewDeployer(
		boshClient,
		serviceAdapter,
		conf.ServiceCatalog,
		conf.BaseDeploymentProperties.Stemcell,
		conf.BaseDeploymentProperties.Releases,
	)

	credHubClient := credhubclient.Client{}
	credhubEnabled := false

	if conf.Credhub != nil {
		credhubEnabled = true
		credHubClient = credhubclient.NewCredhubClient(
			conf.Credhub.APIURL,
			conf.Credhub.ID,
			conf.Credhub.Secret,
			conf.Broker.DisableSSLCertVerification,
		)
	}

	credStore := broker.NewCredentialStore(credhubEnabled, &credHubClient)

	onDemandBroker, err := broker.New(boshClient, cfClient, credStore, serviceAdapter, deploymentManager, conf.ServiceCatalog, loggerFactory)
	if err != nil {
		logger.Fatalf("error starting broker: %s", err)
	}

	if conf.Broker.StartUpBanner {
		fmt.Println(`
                  .//\
        \\      .+ssso/\     \\
      \---.\  .+ssssssso/.  \----\         ____     ______     ______
    .--------+ssssssssssso+--------\      / __ \   (_  __ \   (_   _ \
  .-------:+ssssssssssssssss+--------\   / /  \ \    ) ) \ \    ) (_) )
 -------./ssssssssssssssssssss:.------- ( ()  () )  ( (   ) )   \   _/
  \--------+ssssssssssssssso+--------/  ( ()  () )   ) )  ) )   /  _ \
    \-------.+osssssssssso/.-------/     \ \__/ /   / /__/ /   _) (_) )
      \---./  ./osssssso/   \.---/        \____/   (______/   (______/
        \/      \/osso/       \/
                  \/:/
								`)
	}

	brokerRouter := mux.NewRouter()
	mgmtapi.AttachRoutes(brokerRouter, onDemandBroker, conf.ServiceCatalog, loggerFactory)
	brokerapi.AttachRoutes(brokerRouter, onDemandBroker, lager.NewLogger("on-demand-service-broker"))
	authProtectedBrokerAPI := apiauth.NewWrapper(conf.Broker.Username, conf.Broker.Password).Wrap(brokerRouter)

	// can't use negroni.Classic() if we want custom logger flags in Negroni
	negroniLogger := &negroni.Logger{ALogger: logger}
	server := negroni.New(negroni.NewRecovery(), negroniLogger, negroni.NewStatic(http.Dir("public")))

	server.UseHandler(authProtectedBrokerAPI)
	server.Run(fmt.Sprintf("0.0.0.0:%d", conf.Broker.Port))
}
