package config

import (
	"errors"
	"fmt"
	"io/ioutil"
	"reflect"
	"strings"

	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"

	"gopkg.in/yaml.v2"
)

type Config struct {
	Broker                   Broker
	Credhub                  *Credhub `yaml:",omitempty"`
	Bosh                     Bosh
	CF                       CF
	ServiceAdapter           ServiceAdapter           `yaml:"service_adapter"`
	BaseDeploymentProperties BaseDeploymentProperties `yaml:"service_deployment"`
	ServiceCatalog           broker.ServiceOffering   `yaml:"service_catalog"`
}

func (c Config) Validate() error {
	if err := c.Broker.Validate(); err != nil {
		return err
	}

	if err := c.Bosh.Validate(); err != nil {
		return err
	}

	if err := c.CF.Validate(); err != nil {
		return err
	}

	if c.Credhub != nil {
		if err := c.Credhub.Validate(); err != nil {
			return err
		}
	}

	if err := checkIsExecutableFile(c.ServiceAdapter.Path); err != nil {
		return err
	}

	return nil
}

type Broker struct {
	Port                       int
	Username                   string
	Password                   string
	DisableSSLCertVerification bool `yaml:"disable_ssl_cert_verification"`
	StartUpBanner              bool `yaml:"startup_banner"`
}

func (b Broker) Validate() error {
	if b.Port == 0 {
		return errors.New("broker.port can't be empty")
	}
	if b.Username == "" {
		return errors.New("broker.username can't be empty")
	}
	if b.Password == "" {
		return errors.New("broker.password can't be empty")
	}

	return nil
}

type BaseDeploymentProperties struct {
	Releases serviceadapter.ServiceReleases
	Stemcell serviceadapter.Stemcell
}

type Bosh struct {
	URL            string
	TrustedCert    string `yaml:"root_ca_cert"`
	Authentication BOSHAuthentication
}

type Credhub struct {
	APIURL string `yaml:"api_url"`
	ID     string `yaml:"client_id"`
	Secret string `yaml:"client_secret"`
}

type BOSHAuthentication struct {
	Basic UserCredentials
	UAA   BOSHUAAAuthentication
}

type CF struct {
	URL            string
	TrustedCert    string `yaml:"root_ca_cert"`
	Authentication UAAAuthentication
}

type UserCredentials struct {
	Username string
	Password string
}

func (c UserCredentials) IsSet() bool {
	return c != UserCredentials{}
}

type ClientCredentials struct {
	ID     string `yaml:"client_id"`
	Secret string `yaml:"secret"`
}

type UAAAuthentication struct {
	URL               string            `yaml:"url"`
	ClientCredentials ClientCredentials `yaml:"client_credentials"`
	UserCredentials   UserCredentials   `yaml:"user_credentials"`
}

type BOSHUAAAuthentication struct {
	UAAURL string `yaml:"url"`
	ID     string `yaml:"client_id"`
	Secret string `yaml:"client_secret"`
}

func (cc UAAAuthentication) IsSet() bool {
	return cc != UAAAuthentication{}
}

func (b Bosh) Validate() error {
	if b.URL == "" {
		return fmt.Errorf("Must specify bosh url")
	}
	return b.Authentication.Validate()
}

func (cf CF) Validate() error {
	if cf.URL == "" {
		return fmt.Errorf("Must specify CF url")
	}
	return cf.Authentication.Validate()
}

func (a UAAAuthentication) Validate() error {
	urlIsSet := a.URL != ""
	clientIsSet := a.ClientCredentials.IsSet()
	basicSet := a.UserCredentials.IsSet()

	var err error

	if !urlIsSet && !clientIsSet && !basicSet {
		return fmt.Errorf("Must specify UAA authentication")
	} else if !urlIsSet {
		return fmt.Errorf("Must specify UAA url")
	} else if !clientIsSet && !basicSet {
		return fmt.Errorf("Must specify UAA credentials")
	} else if clientIsSet && basicSet {
		err = fmt.Errorf("Cannot specify both client and user credentials for UAA authentication")
	} else if clientIsSet {
		err = validateNoFieldsEmptyString(a.ClientCredentials, "client_credentials")
	} else if basicSet {
		err = validateNoFieldsEmptyString(a.UserCredentials, "user_credentials")
	}

	return err
}

func (c Credhub) Validate() error {
	if c.APIURL == "" {
		return fmt.Errorf("Must specify a Credhub api_url")
	}

	if c.ID == "" {
		return fmt.Errorf("Must specify a Credhub client_id")
	}

	if c.Secret == "" {
		return fmt.Errorf("Must specify a Credhub client_secret")
	}

	return nil
}

func (cc BOSHUAAAuthentication) IsSet() bool {
	return cc != BOSHUAAAuthentication{}
}

func (cc ClientCredentials) IsSet() bool {
	return cc != ClientCredentials{}
}

func (a BOSHAuthentication) Validate() error {
	uaaSet := a.UAA.IsSet()
	basicSet := a.Basic.IsSet()

	var err error

	if !uaaSet && !basicSet {
		return fmt.Errorf("Must specify bosh authentication")
	} else if uaaSet && basicSet {
		err = fmt.Errorf("Cannot specify both basic and UAA for BOSH authentication")
	} else if uaaSet {
		err = validateNoFieldsEmptyString(a.UAA, "uaa")
	} else if basicSet {
		err = validateNoFieldsEmptyString(a.Basic, "basic")
	}

	return err
}

type ServiceAdapter struct {
	Path string
}

func Parse(configFilePath string) (Config, error) {
	configFileBytes, err := ioutil.ReadFile(configFilePath)
	if err != nil {
		return Config{}, err
	}

	var config Config
	if err := yaml.Unmarshal(configFileBytes, &config); err != nil {
		return Config{}, err
	}

	if err := config.Validate(); err != nil {
		return Config{}, err
	}

	return config, nil
}

func validateNoFieldsEmptyString(obj interface{}, objName string) error {
	bVal := reflect.ValueOf(obj)
	bType := reflect.TypeOf(obj)
	for i := 0; i < bVal.NumField(); i++ {
		fieldVal := bVal.Field(i).String()

		if fieldVal == "" {
			fieldName := bType.Field(i).Name
			return fmt.Errorf("%s.%s can't be empty", objName, strings.ToLower(fieldName))
		}
	}

	return nil
}
