package mgmtapi

import (
	"encoding/json"
	"fmt"
	"io"
	"log"
	"net/http"

	"github.com/gorilla/mux"
	"github.com/pivotal-cf-experimental/service-metrics/metrics"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/loggerfactory"
)

type api struct {
	manageableBroker ManageableBroker
	serviceOffering  broker.ServiceOffering
	loggerFactory    *loggerfactory.LoggerFactory
}

//go:generate counterfeiter -o fake_manageable_broker/fake_manageable_broker.go . ManageableBroker
type ManageableBroker interface {
	Instances(logger *log.Logger) ([]string, error)
	OrphanDeployments(logger *log.Logger) ([]string, error)
	Upgrade(instanceID string, logger *log.Logger) (int, error)
	CountInstancesOfPlans(logger *log.Logger) (map[string]int, error)
}

type Instance struct {
	InstanceID string `json:"instance_id"`
}

type Deployment struct {
	Name string `json:"deployment_name"`
}

type UpgradeResponse struct {
	TaskID int `json:"task_id"`
}

func AttachRoutes(r *mux.Router, manageableBroker ManageableBroker, serviceOffering broker.ServiceOffering, loggerFactory *loggerfactory.LoggerFactory) {
	a := &api{manageableBroker: manageableBroker, serviceOffering: serviceOffering, loggerFactory: loggerFactory}
	r.HandleFunc("/mgmt/service_instances", a.listAllInstances).Methods("GET")
	r.HandleFunc("/mgmt/service_instances/{instance_id}", a.upgradeInstance).Methods("PATCH")
	r.HandleFunc("/mgmt/metrics", a.metrics).Methods("GET")
	r.HandleFunc("/mgmt/orphan_deployments", a.listOrphanDeployments).Methods("GET")
}

func (a *api) listOrphanDeployments(w http.ResponseWriter, r *http.Request) {
	logger := a.loggerFactory.NewLogger()

	orphanNames, err := a.manageableBroker.OrphanDeployments(logger)
	if err != nil {
		logger.Printf("error occurred querying orphan deployments: %s", err)
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	orphanDeployments := []Deployment{}
	for _, name := range orphanNames {
		orphanDeployments = append(orphanDeployments, Deployment{Name: name})
	}

	a.writeJson(w, orphanDeployments, logger)
}

func (a *api) listAllInstances(w http.ResponseWriter, r *http.Request) {
	logger := a.loggerFactory.NewLogger()

	instances, err := a.manageableBroker.Instances(logger)
	if err != nil {
		logger.Printf("error occurred querying instances: %s", err)
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	// now that there is only 1 field in instance, we will just return []string soon, rather than put them into objects
	presentableInstances := []Instance{}
	for _, instance := range instances {
		presentableInstances = append(presentableInstances, Instance{InstanceID: instance})
	}

	a.writeJson(w, presentableInstances, logger)
}

func (a *api) upgradeInstance(w http.ResponseWriter, r *http.Request) {
	logger := a.loggerFactory.NewLogger()

	vars := mux.Vars(r)
	instanceID := vars["instance_id"]
	boshTaskID, err := a.manageableBroker.Upgrade(instanceID, logger)

	switch err.(type) {
	case nil:
		w.WriteHeader(http.StatusAccepted)
		a.writeJson(w, UpgradeResponse{TaskID: boshTaskID}, logger)
	case broker.DeploymentNotFoundError:
		w.WriteHeader(http.StatusGone)
	case broker.OperationInProgressError:
		w.WriteHeader(http.StatusConflict)
	case error:
		logger.Printf("error occurred upgrading instance %s: %s", instanceID, err)
		w.WriteHeader(http.StatusInternalServerError)
		a.writeJson(w, brokerapi.ErrorResponse{Description: err.Error()}, logger)
	}
}

func (a *api) metrics(w http.ResponseWriter, r *http.Request) {
	logger := a.loggerFactory.NewLogger()

	brokerMetrics := []metrics.Metric{}
	instanceCountsByPlan, err := a.manageableBroker.CountInstancesOfPlans(logger)
	if err != nil {
		logger.Printf("error getting instance count for service offering %s: %s", a.serviceOffering.Name, err)
		w.WriteHeader(http.StatusInternalServerError)
		return
	}

	if len(instanceCountsByPlan) == 0 {
		logger.Printf("service %s not registered with Cloud Foundry", a.serviceOffering.Name)
		w.WriteHeader(http.StatusServiceUnavailable)
		return
	}

	for planID, instanceCount := range instanceCountsByPlan {
		plan, err := a.getPlan(planID)
		if err != nil {
			logger.Println(err)
			a.writeJson(w, []interface{}{}, logger)
			return
		}

		countMetric := metrics.Metric{
			Key:   fmt.Sprintf("/on-demand-broker/%s/%s/total_instances", a.serviceOffering.Name, plan.Name),
			Unit:  "count",
			Value: float64(instanceCount),
		}
		brokerMetrics = append(brokerMetrics, countMetric)

		if plan.Quotas != nil {
			quotaMetric := metrics.Metric{
				Key:   fmt.Sprintf("/on-demand-broker/%s/%s/quota_remaining", a.serviceOffering.Name, plan.Name),
				Unit:  "count",
				Value: float64(plan.Quotas.ServiceInstanceLimit - instanceCount),
			}
			brokerMetrics = append(brokerMetrics, quotaMetric)
		}
	}
	a.writeJson(w, brokerMetrics, logger)
}

func (a *api) writeJson(w io.Writer, obj interface{}, logger *log.Logger) {
	if err := json.NewEncoder(w).Encode(obj); err != nil {
		logger.Printf("error occurred encoding json: %s", err)
	}
}

func (a *api) getPlan(planID string) (broker.Plan, error) {
	for _, plan := range a.serviceOffering.Plans {
		if plan.ID == planID {
			return plan, nil
		}
	}
	return broker.Plan{}, fmt.Errorf("no plan found with marketplace ID %s", planID)
}
