package mgmtapi_test

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net/http"
	"net/http/httptest"

	"github.com/gorilla/mux"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/onsi/gomega/gbytes"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/loggerfactory"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi/fake_manageable_broker"
)

var _ = Describe("Management API", func() {
	var (
		server           *httptest.Server
		manageableBroker *fake_manageable_broker.FakeManageableBroker
		logs             *gbytes.Buffer
		loggerFactory    *loggerfactory.LoggerFactory
		serviceOffering  broker.ServiceOffering
	)

	BeforeEach(func() {
		serviceOffering = broker.ServiceOffering{
			ID:    "some_service_offering-id",
			Name:  "some_service_offering",
			Plans: []broker.Plan{{ID: "foo_id", Name: "foo_plan"}},
		}
		logs = gbytes.NewBuffer()
		loggerFactory = loggerfactory.New(io.MultiWriter(GinkgoWriter, logs), "[mgmtapi unit tests] ", log.LstdFlags)
		manageableBroker = new(fake_manageable_broker.FakeManageableBroker)
	})

	JustBeforeEach(func() {
		router := mux.NewRouter()
		mgmtapi.AttachRoutes(router, manageableBroker, serviceOffering, loggerFactory)
		server = httptest.NewServer(router)
	})

	AfterEach(func() {
		server.Close()
	})

	Describe("listing all instances", func() {
		var listResp *http.Response

		JustBeforeEach(func() {
			var err error
			listResp, err = http.Get(fmt.Sprintf("%s/mgmt/service_instances", server.URL))
			Expect(err).NotTo(HaveOccurred())
		})

		Context("of which there are three", func() {
			var (
				instance1 = mgmtapi.Instance{
					InstanceID: "instance-guid-1",
				}
				instance2 = mgmtapi.Instance{
					InstanceID: "instance-guid-2",
				}
				instance3 = mgmtapi.Instance{
					InstanceID: "instance-guid-3",
				}
			)

			BeforeEach(func() {
				instances := []string{instance1.InstanceID, instance2.InstanceID, instance3.InstanceID}
				manageableBroker.InstancesReturns(instances, nil)
			})

			It("returns HTTP 200", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusOK))
			})

			It("returns a list of all three instances", func() {
				var instances []mgmtapi.Instance
				Expect(json.NewDecoder(listResp.Body).Decode(&instances)).To(Succeed())
				Expect(instances).To(ConsistOf(instance1, instance2, instance3))
			})
		})

		Context("but failing to do so", func() {
			BeforeEach(func() {
				manageableBroker.InstancesReturns(nil, errors.New("fnuh"))
			})

			It("returns HTTP 500", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusInternalServerError))
			})

			It("logs the error", func() {
				Eventually(logs).Should(gbytes.Say("error occurred querying instances: fnuh"))
			})
		})
	})

	Describe("upgrading an instance", func() {
		var (
			instanceID = "283974"
			taskID     = 54321

			upgradeResp *http.Response
		)

		JustBeforeEach(func() {
			var err error
			upgradeResp, err = Patch(fmt.Sprintf("%s/mgmt/service_instances/"+instanceID, server.URL))
			Expect(err).NotTo(HaveOccurred())
		})

		Context("successfully", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(taskID, nil)
			})

			It("upgrades the instance using the broker", func() {
				Expect(manageableBroker.UpgradeCallCount()).To(Equal(1))
				actualInstanceID, _ := manageableBroker.UpgradeArgsForCall(0)
				Expect(actualInstanceID).To(Equal(instanceID))
			})

			It("returns HTTP 202", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusAccepted))
			})

			It("returns a bosh task ID", func() {
				var upgradeRespBody mgmtapi.UpgradeResponse
				Expect(json.NewDecoder(upgradeResp.Body).Decode(&upgradeRespBody)).To(Succeed())
				Expect(upgradeRespBody.TaskID).To(Equal(taskID))
			})
		})

		Context("there is a skipable error", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(0, broker.NewDeploymentNotFoundError(fmt.Errorf("this did not work")))
			})

			It("returns HTTP 410", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusGone))
			})
		})

		Context("there is a deployment in progress error", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(0, broker.NewOperationInProgressError(fmt.Errorf("this did not work")))
			})

			It("returns HTTP 409", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusConflict))
			})
		})

		Context("unsuccessfully", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(0, fmt.Errorf("this did not work"))
			})

			It("returns HTTP 500", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusInternalServerError))
			})

			It("returns the error message", func() {
				Expect(ioutil.ReadAll(upgradeResp.Body)).To(MatchJSON(`{"description": "this did not work"}`))
			})

			It("logs the error", func() {
				Eventually(logs).Should(gbytes.Say(fmt.Sprintf("error occurred upgrading instance %s: this did not work", instanceID)))
			})
		})
	})

	Describe("producing service metrics", func() {
		var instancesForPlanResponse *http.Response

		JustBeforeEach(func() {
			var err error
			instancesForPlanResponse, err = http.Get(fmt.Sprintf("%s/mgmt/metrics", server.URL))
			Expect(err).NotTo(HaveOccurred())
		})

		Context("when no quota is set", func() {
			Context("when the instance count can be retrieved", func() {
				BeforeEach(func() {
					manageableBroker.CountInstancesOfPlansReturns(map[string]int{"foo_id": 2}, nil)
				})

				It("returns HTTP 200", func() {
					Expect(instancesForPlanResponse.StatusCode).To(Equal(http.StatusOK))
				})

				It("returns the correct number of instances", func() {
					defer instancesForPlanResponse.Body.Close()
					Expect(ioutil.ReadAll(instancesForPlanResponse.Body)).To(MatchJSON(fmt.Sprintf(`
						[
							{
								"key": "/on-demand-broker/%s/foo_plan/total_instances",
								"unit": "count",
								"value": 2
							}
						]`, serviceOffering.Name)))
				})

				It("counts instances for the plan", func() {
					Expect(manageableBroker.CountInstancesOfPlansCallCount()).To(Equal(1))
				})
			})

			Context("when the instance count cannot be retrieved", func() {
				BeforeEach(func() {
					manageableBroker.CountInstancesOfPlansReturns(nil, errors.New("it done broke"))
				})

				It("returns HTTP 500", func() {
					Expect(instancesForPlanResponse.StatusCode).To(Equal(http.StatusInternalServerError))
				})
			})

			Context("when the broker is not registered with CF", func() {
				BeforeEach(func() {
					manageableBroker.CountInstancesOfPlansReturns(map[string]int{}, nil)
				})

				It("returns HTTP 503", func() {
					Expect(instancesForPlanResponse.StatusCode).To(Equal(http.StatusServiceUnavailable))
				})
			})
		})

		Context("when a quota is set", func() {
			BeforeEach(func() {
				serviceOffering.Plans[0].Quotas = &broker.Quotas{ServiceInstanceLimit: 7}
			})

			Context("when the instance count can be retrieved", func() {
				BeforeEach(func() {
					manageableBroker.CountInstancesOfPlansReturns(map[string]int{"foo_id": 2}, nil)
				})

				It("returns HTTP 200", func() {
					Expect(instancesForPlanResponse.StatusCode).To(Equal(http.StatusOK))
				})

				It("returns the correct number of instances and quota", func() {
					defer instancesForPlanResponse.Body.Close()
					Expect(ioutil.ReadAll(instancesForPlanResponse.Body)).To(MatchJSON(fmt.Sprintf(`
						[
							{
								"key": "/on-demand-broker/%s/foo_plan/total_instances",
								"unit": "count",
								"value": 2
							},
							{
								"key": "/on-demand-broker/%s/foo_plan/quota_remaining",
								"unit": "count",
								"value": 5
							}
						]`, serviceOffering.Name, serviceOffering.Name)))
				})

				It("counts instances for the plan", func() {
					Expect(manageableBroker.CountInstancesOfPlansCallCount()).To(Equal(1))
				})
			})
		})
	})

	Describe("listing orphan service deployments", func() {
		var listResp *http.Response

		JustBeforeEach(func() {
			var err error
			listResp, err = http.Get(fmt.Sprintf("%s/mgmt/orphan_deployments", server.URL))
			Expect(err).NotTo(HaveOccurred())
		})

		Context("when there are no orphans", func() {
			It("returns HTTP 200", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusOK))
			})

			It("returns no deployments", func() {
				defer listResp.Body.Close()
				body, err := ioutil.ReadAll(listResp.Body)
				Expect(err).NotTo(HaveOccurred())
				Expect(body).To(MatchJSON(`[]`))
			})
		})

		Context("when there are some orphans", func() {
			var (
				orphan1 = mgmtapi.Deployment{
					Name: "orphan1",
				}
				orphan2 = mgmtapi.Deployment{
					Name: "orphan2",
				}
			)

			BeforeEach(func() {
				manageableBroker.OrphanDeploymentsReturns([]string{orphan1.Name, orphan2.Name}, nil)
			})

			It("returns HTTP 200", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusOK))
			})

			It("returns some deployments", func() {
				var orphans []mgmtapi.Deployment
				Expect(json.NewDecoder(listResp.Body).Decode(&orphans)).To(Succeed())
				Expect(orphans).To(ConsistOf(orphan1, orphan2))
			})
		})

		Context("when broker returns an error", func() {
			BeforeEach(func() {
				manageableBroker.OrphanDeploymentsReturns([]string{}, errors.New("Broker errored."))
			})

			It("returns HTTP 500", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusInternalServerError))
			})

			It("returns an empty body", func() {
				defer listResp.Body.Close()
				body, err := ioutil.ReadAll(listResp.Body)
				Expect(err).NotTo(HaveOccurred())
				Expect(body).To(BeEmpty())
			})

			It("logs an error", func() {
				Eventually(logs).Should(gbytes.Say("error occurred querying orphan deployments: Broker errored."))
			})
		})
	})
})

func Patch(url string) (resp *http.Response, err error) {
	req, err := http.NewRequest("PATCH", url, nil)
	if err != nil {
		return nil, err
	}
	return http.DefaultClient.Do(req)
}
