package upgrader

import (
	"encoding/json"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
)

type upgrader struct {
	brokerUsername  string
	brokerPassword  string
	brokerUrl       string
	pollingInterval int64
}

func New(brokerUsername, brokerPassword, brokerUrl string, pollingInterval int64) upgrader {
	return upgrader{
		brokerUsername:  brokerUsername,
		brokerPassword:  brokerPassword,
		brokerUrl:       brokerUrl,
		pollingInterval: pollingInterval,
	}
}

func (u upgrader) Upgrade() (upgraded []string, skipped []string, err error) {
	errs := func(err error) ([]string, []string, error) {
		return nil, nil, err
	}

	instances, err := u.getServiceInstances()
	if err != nil {
		return errs(fmt.Errorf("error getting service instances: %s\n", err))
	}
	for len(instances) != 0 {
		upgradedInstances, skippedInstances, retryInstances, err := u.upgradeInstances(instances)
		if err != nil {
			return errs(err)
		}
		upgraded = append(upgraded, upgradedInstances...)
		skipped = append(skipped, skippedInstances...)
		instances = retryInstances

		if len(instances) > 0 {
			u.sleep()
		}
	}
	return
}

func (u upgrader) upgradeInstances(instances []string) (upgradedInstances []string, skippedInstances []string, retryInstances []string, err error) {
	errs := func(err error) ([]string, []string, []string, error) {
		return nil, nil, nil, err
	}

	for _, instance := range instances {
		taskID, skip, retry, err := u.upgradeInstance(instance)
		if err != nil {
			return errs(fmt.Errorf("Upgrade failed for service_instance_id %s: %s\n", instance, err))
		}

		if skip {
			skippedInstances = append(skippedInstances, instance)
			continue
		}
		if retry {
			retryInstances = append(retryInstances, instance)
			continue
		}

		if err := u.pollLastOperation(instance, taskID); err != nil {
			return errs(err)
		}

		upgradedInstances = append(upgradedInstances, instance)
	}
	return
}

func (u upgrader) sleep() {
	time.Sleep(time.Millisecond * time.Duration(u.pollingInterval))
}

func (u upgrader) pollLastOperation(instance string, taskID int) error {
	lastOperation := brokerapi.LastOperation{State: brokerapi.InProgress}
	var err error
	for lastOperation.State == brokerapi.InProgress {
		u.sleep()
		lastOperation, err = u.getLastOperation(instance, taskID)
		if err != nil {
			return fmt.Errorf("error getting last operation: %s\n", err)
		}
	}

	if lastOperation.State != brokerapi.Succeeded {
		return fmt.Errorf("Upgrade failed for service_instance_id %s: %s", instance, lastOperation.Description)
	}
	return nil
}

func (u upgrader) getServiceInstances() ([]string, error) {
	var instances []mgmtapi.Instance
	if err := u.makeRequest("/mgmt/service_instances", &instances); err != nil {
		return nil, err
	}

	var instanceIDs []string
	for _, instance := range instances {
		instanceIDs = append(instanceIDs, instance.InstanceID)
	}
	return instanceIDs, nil
}

func (u upgrader) upgradeInstance(instance string) (taskID int, skip bool, retry bool, err error) {
	request, err := http.NewRequest("PATCH", fmt.Sprintf("%s%s", u.brokerUrl, "/mgmt/service_instances/"+instance), nil)
	if err != nil {
		return 0, false, false, err
	}
	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return 0, false, false, err
	}
	defer response.Body.Close()

	switch response.StatusCode {
	case http.StatusAccepted:
		var upgradeResp mgmtapi.UpgradeResponse
		if err := json.NewDecoder(response.Body).Decode(&upgradeResp); err != nil {
			return 0, false, false, fmt.Errorf("cant parse response %s", err)
		}

		return upgradeResp.TaskID, false, false, nil
	case http.StatusGone:
		return 0, true, false, nil
	case http.StatusConflict:
		return 0, false, true, nil
	default:
		var errorResponse brokerapi.ErrorResponse
		if err := json.NewDecoder(response.Body).Decode(&errorResponse); err != nil {
			return 0, false, false, fmt.Errorf("cant parse response %s", err)
		}
		return 0, false, false, fmt.Errorf("ERROR: %s", errorResponse.Description)
	}
}

func (u upgrader) getLastOperation(instance string, taskID int) (brokerapi.LastOperation, error) {
	operationData, err := json.Marshal(broker.OperationData{BoshTaskID: taskID, OperationType: broker.OperationTypeUpgrade})
	if err != nil {
		return brokerapi.LastOperation{}, err
	}
	operationQueryStr := url.QueryEscape(string(operationData))

	var lastOperation brokerapi.LastOperation
	if err := u.makeRequest("/v2/service_instances/"+instance+"/last_operation?operation="+operationQueryStr, &lastOperation); err != nil {
		return brokerapi.LastOperation{}, err
	}
	return lastOperation, nil
}

func (u upgrader) makeRequest(requestPath string, responseData interface{}) error {
	request, err := http.NewRequest("GET", fmt.Sprintf("%s%s", u.brokerUrl, requestPath), nil)
	if err != nil {
		return err
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return err
	}

	if response.StatusCode != 200 {
		return fmt.Errorf("response returned bad status: %s", response.Status)
	}

	err = json.NewDecoder(response.Body).Decode(responseData)
	if err != nil {
		return err
	}

	return nil
}
